# -*- coding: utf-8 -*-

"""
This module helps you convert an AST back to readable Python code.

The module is adapted from https://github.com/JelleZijlstra/ast_decompiler.
"""

import ast
import sys
from contextlib import contextmanager

__all__ = [
  'ast2code',
]


def ast2code(ast_node, indent=4, line_length=300):
  """Decompiles an AST into Python code.

  Parameters
  ----------
  ast_node : ast.Node
      Code to decompile, using AST objects as generated by the standard library ast module.
  indent : int
      Indent level of lines.
  line_length : int
      If the line become longer than "line_length", this module will break them up

  Returns
  -------
  code : str
      Python code string.
  """
  decompiler = Transformer(indentation=indent,
                           line_length=line_length, )
  return decompiler.run(ast_node)


class _CallArgs(object):
  def __init__(self, args):
    self.args = args


class KeyValuePair(object):
  """A key-value pair as used in a dictionary display."""
  _fields = ['key', 'value']

  def __init__(self, key, value):
    self.key = key
    self.value = value


class StarArg(object):
  """A * argument."""
  _fields = ['arg']

  def __init__(self, arg):
    self.arg = arg


class DoubleStarArg(object):
  """A ** argument."""
  _fields = ['arg']

  def __init__(self, arg):
    self.arg = arg


class KeywordArg(object):
  """A x=3 keyword argument in a function definition."""
  _fields = ['arg', 'value']

  def __init__(self, arg, value):
    self.arg = arg
    self.value = value


_OP_TO_STR = {
  ast.Add: '+',
  ast.Sub: '-',
  ast.Mult: '*',
  ast.Div: '/',
  ast.Mod: '%',
  ast.Pow: '**',
  ast.LShift: '<<',
  ast.RShift: '>>',
  ast.BitOr: '|',
  ast.BitXor: '^',
  ast.BitAnd: '&',
  ast.FloorDiv: '//',
  ast.Invert: '~',
  ast.Not: 'not ',
  ast.UAdd: '+',
  ast.USub: '-',
  ast.Eq: '==',
  ast.NotEq: '!=',
  ast.Lt: '<',
  ast.LtE: '<=',
  ast.Gt: '>',
  ast.GtE: '>=',
  ast.Is: 'is',
  ast.IsNot: 'is not',
  ast.In: 'in',
  ast.NotIn: 'not in',
  ast.And: 'and',
  ast.Or: 'or',
}

_PRECEDENCE = {
  _CallArgs: -1,
  ast.Or: 0,
  ast.And: 1,
  ast.Not: 2,
  ast.Compare: 3,
  ast.BitOr: 4,
  ast.BitXor: 5,
  ast.BitAnd: 6,
  ast.LShift: 7,
  ast.RShift: 7,
  ast.Add: 8,
  ast.Sub: 8,
  ast.Mult: 9,
  ast.Div: 9,
  ast.FloorDiv: 9,
  ast.Mod: 9,
  ast.UAdd: 10,
  ast.USub: 10,
  ast.Invert: 10,
  ast.Pow: 11,
  ast.Subscript: 12,
  ast.Call: 12,
  ast.Attribute: 12,
}

if hasattr(ast, 'MatMult'):
  _OP_TO_STR[ast.MatMult] = '@'
  _PRECEDENCE[ast.MatMult] = 9


class Transformer(ast.NodeVisitor):
  def __init__(self, indentation, line_length):
    self.lines = []
    self.current_line = []
    self.current_indentation = 0
    self.node_stack = []
    self.indentation = indentation
    self.max_line_length = line_length
    self.has_unicode_literals = False

  def run(self, ast):
    self.visit(ast)
    if self.current_line:
      self.lines.append(''.join(self.current_line))
      self.current_line = []
    return ''.join(self.lines)

  def visit(self, node):
    self.node_stack.append(node)
    try:
      return super(Transformer, self).visit(node)
    finally:
      if self.node_stack:
        self.node_stack.pop()

  def precedence_of_node(self, node):
    if isinstance(node, (ast.BinOp, ast.UnaryOp, ast.BoolOp)):
      return _PRECEDENCE[type(node.op)]
    return _PRECEDENCE.get(type(node), -1)

  def get_parent_node(self):
    try:
      return self.node_stack[-2]
    except IndexError:
      return None

  def has_parent_of_type(self, node_type):
    return any(isinstance(parent, node_type) for parent in self.node_stack)

  def write(self, code):
    assert isinstance(code, str), 'invalid code %r' % code
    self.current_line.append(code)

  def write_indentation(self):
    self.write(' ' * self.current_indentation)

  def write_newline(self):
    line = ''.join(self.current_line) + '\n'
    self.lines.append(line)
    self.current_line = []

  def current_line_length(self):
    return sum(map(len, self.current_line))

  def write_expression_list(self, nodes, separator=', ', allow_newlines=True, need_parens=True,
                            final_separator_if_multiline=True):
    """Write a list of nodes, separated by separator.

    Parameters
    ----------
    nodes : list, tuple
        Nodes.
    separator : str
        Separators.
    allow_newlines : bool
        If true, this function will write_module the expression over multiple lines.
    need_parens : bool
        If true, this function will surround the expression with parentheses in this case.
    final_separator_if_multiline : bool
        If true, this function will write_module a separator at the end of the list if it is
        divided over multiple lines.
    """
    first = True
    last_line = len(self.lines)
    current_line = list(self.current_line)
    for node in nodes:
      if first:
        first = False
      else:
        self.write(separator)
      self.visit(node)
      if allow_newlines and (self.current_line_length() > self.max_line_length or
                             last_line != len(self.lines)):
        break
    else:
      return

    # reset state
    del self.lines[last_line:]
    self.current_line = current_line

    separator = separator.rstrip()
    if need_parens:
      self.write('(')
    self.write_newline()
    with self.add_indentation():
      num_nodes = len(nodes)
      for i, node in enumerate(nodes):
        self.write_indentation()
        self.visit(node)
        if final_separator_if_multiline or i < num_nodes - 1:
          self.write(separator)
        self.write_newline()

    self.write_indentation()
    if need_parens:
      self.write(')')

  def write_suite(self, nodes):
    with self.add_indentation():
      for line in nodes:
        self.visit(line)

  @contextmanager
  def add_indentation(self):
    self.current_indentation += self.indentation
    try:
      yield
    finally:
      self.current_indentation -= self.indentation

  @contextmanager
  def parenthesize_if(self, condition):
    if condition:
      self.write('(')
      yield
      self.write(')')
    else:
      yield

  @contextmanager
  def f_literalise_if(self, condition):
    if condition:
      self.write("f'")
      yield
      self.write("'")
    else:
      yield

  def generic_visit(self, node):
    raise NotImplementedError('missing visit method for %r' % node)

  def visit_Module(self, node):
    for line in node.body:
      self.visit(line)

  visit_Interactive = visit_Module

  def visit_Expression(self, node):
    self.visit(node.body)

  # Multi-line statements

  def visit_FunctionDef(self, node):
    self.write_function_def(node)

  def visit_AsyncFunctionDef(self, node):
    self.write_function_def(node, is_async=True)

  def write_function_def(self, node, is_async=False):
    self.write_newline()
    for decorator in node.decorator_list:
      self.write_indentation()
      self.write('@')
      self.visit(decorator)
      self.write_newline()

    self.write_indentation()
    if is_async:
      self.write('async ')
    self.write('def %s(' % node.name)
    self.visit(node.args)
    self.write(')')
    if getattr(node, 'returns', None):
      self.write(' -> ')
      self.visit(node.return_intermediates)
    self.write(':')
    self.write_newline()

    self.write_suite(node.body)

  def visit_ClassDef(self, node):
    self.write_newline()
    self.write_newline()
    for decorator in node.decorator_list:
      self.write_indentation()
      self.write('@')
      self.visit(decorator)
      self.write_newline()

    self.write_indentation()
    self.write('class %s(' % node.name)
    exprs = node.bases + getattr(node, 'keywords', [])
    self.write_expression_list(exprs, need_parens=False)
    self.write('):')
    self.write_newline()
    self.write_suite(node.body)

  def visit_For(self, node):
    self.write_for(node)

  def visit_AsyncFor(self, node):
    self.write_for(node, is_async=True)

  def write_for(self, node, is_async=False):
    self.write_indentation()
    if is_async:
      self.write('async ')
    self.write('for ')
    self.visit(node.target)
    self.write(' in ')
    self.visit(node.iter)
    self.write(':')
    self.write_newline()
    self.write_suite(node.body)
    self.write_else(node.orelse)

  def visit_While(self, node):
    self.write_indentation()
    self.write('while ')
    self.visit(node.test)
    self.write(':')
    self.write_newline()
    self.write_suite(node.body)
    self.write_else(node.orelse)

  def visit_If(self, node):
    self.write_indentation()
    self.write('if ')
    self.visit(node.test)
    self.write(':')
    self.write_newline()
    self.write_suite(node.body)
    while node.orelse and len(node.orelse) == 1 and isinstance(node.orelse[0], ast.If):
      node = node.orelse[0]
      self.write_indentation()
      self.write('elif ')
      self.visit(node.test)
      self.write(':')
      self.write_newline()
      self.write_suite(node.body)
    self.write_else(node.orelse)

  def write_else(self, orelse):
    if orelse:
      self.write_indentation()
      self.write('else:')
      self.write_newline()
      self.write_suite(orelse)

  def visit_With(self, node):
    if hasattr(node, 'items'):
      self.write_py3_with(node)
      return
    self.write_indentation()
    self.write('with ')
    nodes = [node]
    body = node.body
    is_first = True

    while len(body) == 1 and isinstance(body[0], ast.With):
      nodes.append(body[0])
      body = body[0].body

    for context_node in nodes:
      if is_first:
        is_first = False
      else:
        self.write(', ')
      self.visit(context_node.context_expr)
      if context_node.optional_vars:
        self.write(' as ')
        self.visit(context_node.optional_vars)
    self.write(':')
    self.write_newline()
    self.write_suite(body)

  def visit_AsyncWith(self, node):
    self.write_py3_with(node, is_async=True)

  def write_py3_with(self, node, is_async=False):
    self.write_indentation()
    if is_async:
      self.write('async ')
    self.write('with ')
    self.write_expression_list(node.items, allow_newlines=False)
    self.write(':')
    self.write_newline()
    self.write_suite(node.body)

  def visit_withitem(self, node):
    self.visit(node.context_expr)
    if node.optional_vars:
      self.write(' as ')
      self.visit(node.optional_vars)

  def visit_Try(self, node):
    # py3 only
    self.visit_TryExcept(node)
    if node.finalbody:
      self.write_finalbody(node.finalbody)

  def visit_TryExcept(self, node):
    self.write_indentation()
    self.write('try:')
    self.write_newline()
    self.write_suite(node.body)
    for handler in node.handlers:
      self.visit(handler)
    self.write_else(node.orelse)

  def visit_TryFinally(self, node):
    if len(node.body) == 1 and isinstance(node.body[0], ast.Try):
      self.visit(node.body[0])
    else:
      self.write_indentation()
      self.write('try:')
      self.write_newline()
      self.write_suite(node.body)
    self.write_finalbody(node.finalbody)

  def write_finalbody(self, body):
    self.write_indentation()
    self.write('finally:')
    self.write_newline()
    self.write_suite(body)

  # One-line statements

  def visit_Return(self, node):
    self.write_indentation()
    self.write('return')
    if node.value:
      self.write(' ')
      self.visit(node.value)
    self.write_newline()

  def visit_Delete(self, node):
    self.write_indentation()
    self.write('del ')
    self.write_expression_list(node.targets, allow_newlines=False)
    self.write_newline()

  def visit_Assign(self, node):
    self.write_indentation()
    self.write_expression_list(node.targets, separator=' = ', allow_newlines=False)
    self.write(' = ')
    self.visit(node.value)
    self.write_newline()

  def visit_AugAssign(self, node):
    self.write_indentation()
    self.visit(node.target)
    self.write(' ')
    self.visit(node.op)
    self.write('= ')
    self.visit(node.value)
    self.write_newline()

  def visit_AnnAssign(self, node):
    self.write_indentation()
    if not node.simple:
      self.write('(')
    self.visit(node.target)
    if not node.simple:
      self.write(')')
    self.write(': ')
    self.visit(node.annotation)
    if node.value is not None:
      self.write(' = ')
      self.visit(node.value)
    self.write_newline()

  def visit_Print(self, node):
    self.write_indentation()
    self.write('print')
    if node.dest:
      self.write(' >>')
      self.visit(node.dest)
      if node.values:
        self.write(',')
    if node.values:
      self.write(' ')
    self.write_expression_list(node.values, allow_newlines=False)
    if not node.nl:
      self.write(',')
    self.write_newline()

  def visit_Raise(self, node):
    self.write_indentation()
    self.write('raise')
    if hasattr(node, 'exc'):
      # py3
      if node.exc is not None:
        self.write(' ')
        self.visit(node.exc)
        if node.cause is not None:
          self.write(' from ')
          self.visit(node.cause)
    else:
      expressions = [child for child in (node.type, node.inst, node.tback) if child]
      if expressions:
        self.write(' ')
        self.write_expression_list(expressions, allow_newlines=False)
    self.write_newline()

  def visit_Assert(self, node):
    self.write_indentation()
    self.write('assert ')
    self.visit(node.test)
    if node.msg:
      self.write(', ')
      self.visit(node.msg)
    self.write_newline()

  def visit_Import(self, node):
    self.write_indentation()
    self.write('import ')
    self.write_expression_list(node.names, allow_newlines=False)
    self.write_newline()

  def visit_ImportFrom(self, node):
    if node.module == '__future__' and any(alias.name == 'unicode_literals' for alias in node.names):
      self.has_unicode_literals = True

    self.write_indentation()
    self.write('from %s' % ('.' * (node.level or 0)))
    if node.module:
      self.write(node.module)
    self.write(' import ')
    self.write_expression_list(node.names)
    self.write_newline()

  def visit_Exec(self, node):
    self.write_indentation()
    self.write('exec ')
    self.visit(node.body)
    if node.globals:
      self.write(' in ')
      self.visit(node.globals)
    if node.locals:
      self.write(', ')
      self.visit(node.locals)
    self.write_newline()

  def visit_Global(self, node):
    self.write_indentation()
    self.write('global %s' % ', '.join(node.names))
    self.write_newline()

  def visit_Nonlocal(self, node):
    self.write_indentation()
    self.write('nonlocal %s' % ', '.join(node.names))
    self.write_newline()

  def visit_Expr(self, node):
    self.write_indentation()
    self.visit(node.value)
    self.write_newline()

  def visit_Pass(self, node):
    self.write_indentation()
    self.write('pass')
    self.write_newline()

  def visit_Break(self, node):
    self.write_indentation()
    self.write('break')
    self.write_newline()

  def visit_Continue(self, node):
    self.write_indentation()
    self.write('continue')
    self.write_newline()

  # Expressions

  def visit_BoolOp(self, node):
    my_prec = self.precedence_of_node(node)
    parent_prec = self.precedence_of_node(self.get_parent_node())
    with self.parenthesize_if(my_prec <= parent_prec):
      op = 'and' if isinstance(node.op, ast.And) else 'or'
      self.write_expression_list(
        node.values,
        separator=' %s ' % op,
        final_separator_if_multiline=False,
      )

  def visit_BinOp(self, node):
    parent_node = self.get_parent_node()
    my_prec = self.precedence_of_node(node)
    parent_prec = self.precedence_of_node(parent_node)
    if my_prec < parent_prec:
      should_parenthesize = True
    elif my_prec == parent_prec:
      if isinstance(node.op, ast.Pow):
        should_parenthesize = node == parent_node.left
      else:
        should_parenthesize = node == parent_node.right
    else:
      should_parenthesize = False

    with self.parenthesize_if(should_parenthesize):
      self.visit(node.left)
      self.write(' ')
      self.visit(node.op)
      self.write(' ')
      self.visit(node.right)

  def visit_UnaryOp(self, node):
    my_prec = self.precedence_of_node(node)
    parent_prec = self.precedence_of_node(self.get_parent_node())
    with self.parenthesize_if(my_prec < parent_prec):
      self.visit(node.op)
      self.visit(node.operand)

  def visit_Lambda(self, node):
    should_parenthesize = isinstance(
      self.get_parent_node(),
      (ast.BinOp, ast.UnaryOp, ast.Compare, ast.IfExp, ast.Attribute, ast.Subscript, ast.Call, ast.BoolOp)
    )
    with self.parenthesize_if(should_parenthesize):
      self.write('lambda')
      if node.args.args or node.args.vararg or node.args.kwarg:
        self.write(' ')
      self.visit(node.args)
      self.write(': ')
      self.visit(node.body)

  def visit_NamedExpr(self, node):
    self.write('(')
    self.visit(node.target)
    self.write(' := ')
    # := has the lowest precedence, so we should never need to parenthesize this
    self.visit(node.value)
    self.write(')')

  def visit_IfExp(self, node):
    parent_node = self.get_parent_node()
    if isinstance(
        parent_node,
        (ast.BinOp, ast.UnaryOp, ast.Compare, ast.Attribute, ast.Subscript,
         ast.Call, ast.BoolOp, ast.comprehension)
    ):
      should_parenthesize = True
    elif isinstance(parent_node, ast.IfExp) and (node is parent_node.test or node is parent_node.body):
      should_parenthesize = True
    else:
      should_parenthesize = False

    with self.parenthesize_if(should_parenthesize):
      self.visit(node.body)
      self.write(' if ')
      self.visit(node.test)
      self.write(' else ')
      self.visit(node.orelse)

  def visit_Dict(self, node):
    self.write('{')
    items = [KeyValuePair(key, value) for key, value in zip(node.keys, node.values)]
    self.write_expression_list(items, need_parens=False)
    self.write('}')

  def visit_KeyValuePair(self, node):
    self.visit(node.key)
    self.write(': ')
    self.visit(node.value)

  def visit_Set(self, node):
    self.write('{')
    self.write_expression_list(node.elts, need_parens=False)
    self.write('}')

  def visit_ListComp(self, node):
    self.visit_comp(node, '[', ']')

  def visit_SetComp(self, node):
    self.visit_comp(node, '{', '}')

  def visit_DictComp(self, node):
    self.write('{')
    elts = [KeyValuePair(node.key, node.value)] + node.generators
    self.write_expression_list(elts, separator=' ', need_parens=False)
    self.write('}')

  def visit_GeneratorExp(self, node):
    parent_node = self.get_parent_node()
    # if this is the only argument to a function, omit the extra parentheses
    if (isinstance(parent_node, _CallArgs) and len(parent_node.args) == 1 and node == parent_node.args[0]):
      start = end = ''
    else:
      start = '('
      end = ')'
    self.visit_comp(node, start, end)

  def visit_comp(self, node, start, end):
    self.write(start)
    self.write_expression_list([node.elt] + node.generators, separator=' ', need_parens=False)
    self.write(end)

  def visit_Await(self, node):
    with self.parenthesize_if(not isinstance(self.get_parent_node(), (ast.Expr, ast.Assign, ast.AugAssign))):
      self.write('await ')
      self.visit(node.value)

  def visit_Yield(self, node):
    with self.parenthesize_if(not isinstance(self.get_parent_node(), (ast.Expr, ast.Assign, ast.AugAssign))):
      self.write('yield')
      if node.value:
        self.write(' ')
        self.visit(node.value)

  def visit_YieldFrom(self, node):
    with self.parenthesize_if(not isinstance(self.get_parent_node(), (ast.Expr, ast.Assign, ast.AugAssign))):
      self.write('yield from ')
      self.visit(node.value)

  def visit_Compare(self, node):
    my_prec = self.precedence_of_node(node)
    parent_prec = self.precedence_of_node(self.get_parent_node())
    with self.parenthesize_if(my_prec <= parent_prec):
      self.visit(node.left)
      for op, expr in zip(node.ops, node.comparators):
        self.write(' ')
        self.visit(op)
        self.write(' ')
        self.visit(expr)

  def visit_Call(self, node):
    self.visit(node.func)
    self.write('(')

    args = node.args + node.keywords
    if hasattr(node, 'starargs') and node.starargs:
      args.append(StarArg(node.starargs))
    if hasattr(node, 'kwargs') and node.kwargs:
      args.append(DoubleStarArg(node.kwargs))
    self.node_stack.append(_CallArgs(args))
    try:
      if args:
        self.write_expression_list(
          args,
          need_parens=False,
          final_separator_if_multiline=False  # it's illegal after *args and **kwargs
        )

      self.write(')')
    finally:
      self.node_stack.pop()

  def visit_StarArg(self, node):
    self.write('*')
    self.visit(node.arg)

  def visit_DoubleStarArg(self, node):
    self.write('**')
    self.visit(node.arg)

  def visit_KeywordArg(self, node):
    self.visit(node.arg)
    if node.value is not None:
      self.write('=')
      self.visit(node.value)

  def visit_Repr(self, node):
    self.write('`')
    self.visit(node.value)
    self.write('`')

  def visit_Num(self, node):
    self.write_number(node.n)

  def write_number(self, number):
    should_parenthesize = isinstance(number, int) and number >= 0 and \
                          isinstance(self.get_parent_node(), ast.Attribute)
    should_parenthesize = should_parenthesize or \
                          (isinstance(number, complex) and
                           number.real == 0.0 and
                           (number.imag < 0 or number.imag == -0.0))
    # Always parenthesize in Python 2, because there "-(1)" and "-1" produce a different AST.
    if not should_parenthesize and (isinstance(number, complex) or number < 0):
      parent_node = self.get_parent_node()
      should_parenthesize = isinstance(parent_node, ast.UnaryOp) and \
                            isinstance(parent_node.op, ast.USub) and \
                            hasattr(parent_node, 'lineno')
    with self.parenthesize_if(should_parenthesize):
      if isinstance(number, float) and abs(number) > sys.float_info.max:
        # otherwise we write inf, which won't be parsed back right
        # I don't know of any way to write nan with a literal
        self.write('1e1000' if number > 0 else '-1e1000')
      elif isinstance(number, (int, int, float)) and number < 0:
        # needed for precedence to work correctly
        me = self.node_stack.pop()
        if isinstance(number, int):
          val = str(-number)
        else:
          val = repr(type(number)(-number))  # - of long may be int
        self.visit(ast.UnaryOp(op=ast.USub(), operand=ast.Name(id=val)))
        self.node_stack.append(me)
      else:
        self.write(repr(number))

  def visit_Str(self, node):
    self.write_string(node.s, kind=None)

  def write_string(self, string_value, kind=None):
    if sys.version_info >= (3, 6) and self.has_parent_of_type(ast.FormattedValue):
      delimiter = '"'
    else:
      delimiter = "'"
    self.write(delimiter)
    s = string_value.encode('unicode-escape').decode('ascii')
    self.write(s.replace(delimiter, '\\' + delimiter))
    self.write(delimiter)

  def visit_FormattedValue(self, node):
    has_parent = isinstance(self.get_parent_node(), ast.JoinedStr)
    with self.f_literalise_if(not has_parent):
      self.write('{')
      if isinstance(node.value, ast.JoinedStr):
        raise NotImplementedError('ast_decompiler does not support nested f-strings yet')
      add_space = isinstance(node.value, (ast.Set, ast.Dict, ast.SetComp, ast.DictComp))
      if add_space:
        self.write(' ')
      self.visit(node.value)
      if node.conversion != -1:
        self.write('!%s' % chr(node.conversion))
      if node.format_spec is not None:
        self.write(':')
        if isinstance(node.format_spec, ast.JoinedStr):
          self.visit(node.format_spec)
        elif isinstance(node.format_spec, ast.Str):
          self.write(node.format_spec.s)
        else:
          raise TypeError('format spec must be a string, not {}'.format(node.format_spec))
      if add_space:
        self.write(' ')
      self.write('}')

  def visit_JoinedStr(self, node):
    has_parent = isinstance(self.get_parent_node(), ast.FormattedValue)
    with self.f_literalise_if(not has_parent):
      for value in node.values:
        if isinstance(value, ast.Str):
          # always escape '
          self.write(value.s.encode('unicode-escape').decode('ascii').replace("'", r"\'"))
        else:
          self.visit(value)

  def visit_Bytes(self, node):
    self.write(repr(node.s))

  def visit_NameConstant(self, node):
    self.write(repr(node.value))

  def visit_Constant(self, node):
    if node.value is Ellipsis:
      self.write('...')
    elif isinstance(node.value, str):
      self.write_string(node.value, node.kind)
    elif isinstance(node.value, bytes):
      self.write(repr(node.value))
    elif isinstance(node.value, (int, float, complex)):
      self.write_number(node.value)
    elif isinstance(node.value, (bool, type(None))):
      self.write(repr(node.value))
    else:
      raise NotImplementedError(ast.dump(node))

  def visit_Attribute(self, node):
    self.visit(node.value)
    self.write('.%s' % node.attr)

  def visit_Subscript(self, node):
    self.visit(node.value)
    self.write('[')
    self.visit(node.slice)
    self.write(']')

  def visit_Starred(self, node):
    self.write('*')
    self.visit(node.value)

  def visit_Name(self, node):
    if isinstance(node.id, str):
      self.write(node.id)
    else:
      self.visit(node.id)

  def visit_List(self, node):
    self.write('[')
    self.write_expression_list(node.elts, need_parens=False)
    self.write(']')

  def visit_Tuple(self, node):
    if not node.elts:
      self.write('()')
    else:
      parent_node = self.get_parent_node()
      should_parenthesize = not isinstance(
        parent_node,
        (ast.Expr, ast.Assign, ast.AugAssign, ast.Return, ast.Yield)
      )
      if isinstance(parent_node, ast.comprehension) and node is parent_node.target:
        should_parenthesize = False
      # https://bugs.python.org/issue32117
      if (
          hasattr(ast, 'Starred')
          and isinstance(parent_node, (ast.Return, ast.Yield))
          and any(isinstance(elt, ast.Starred) for elt in node.elts)
          and sys.version_info < (3, 8)
      ):
        should_parenthesize = True
      with self.parenthesize_if(should_parenthesize):
        if len(node.elts) == 1:
          self.visit(node.elts[0])
          self.write(',')
        else:
          self.write_expression_list(node.elts, need_parens=not should_parenthesize)

  # slice

  def visit_Ellipsis(self, node):
    self.write('...')

  def visit_Slice(self, node):
    if node.lower:
      self.visit(node.lower)
    self.write(':')
    if node.upper:
      self.visit(node.upper)
    if node.step:
      self.write(':')
      self.visit(node.step)

  def visit_ExtSlice(self, node):
    if len(node.dims) == 1:
      self.visit(node.dims[0])
      self.write(',')
    else:
      self.write_expression_list(node.dims, need_parens=False)

  def visit_Index(self, node):
    self.visit(node.value)

  # operators
  for op, string in _OP_TO_STR.items():
    exec('def visit_%s(self, node): self.write(%r)' % (op.__name__, string))

  # Other types

  visit_Load = visit_Store = visit_Del = visit_AugLoad = visit_AugStore = visit_Param = \
    lambda self, node: None

  def visit_comprehension(self, node):
    if getattr(node, "is_async", False):
      self.write('async ')
    self.write('for ')
    self.visit(node.target)
    self.write(' in ')
    self.visit(node.iter)
    for expr in node.ifs:
      self.write(' if ')
      self.visit(expr)

  def visit_ExceptHandler(self, node):
    self.write_indentation()
    self.write('except')
    if node.type:
      self.write(' ')
      self.visit(node.type)
      if node.name:
        self.write(' as ')
        # node.name is a string in py3 but an expr in py2
        if isinstance(node.name, str):
          self.write(node.name)
        else:
          self.visit(node.name)
    self.write(':')
    self.write_newline()
    self.write_suite(node.body)

  def visit_arguments(self, node):
    args = []
    if hasattr(node, "posonlyargs") and node.posonlyargs:
      args += node.posonlyargs
      args.append(ast.Name(id='/'))

    num_defaults = len(node.defaults)
    if num_defaults:
      args += node.args[:-num_defaults]
      default_args = zip(node.args[-num_defaults:], node.defaults)
    else:
      args += list(node.args)
      default_args = []
    for name, value in default_args:
      args.append(KeywordArg(name, value))

    if node.vararg:
      args.append(StarArg(ast.Name(id=node.vararg)))

    if hasattr(node, 'kw_defaults'):
      if node.kwonlyargs and not node.vararg:
        args.append(StarArg(ast.Name(id='')))
      num_kwarg_defaults = len(node.kw_defaults)
      if num_kwarg_defaults:
        args += node.kwonlyargs[:-num_kwarg_defaults]
        default_kwargs = zip(node.kwonlyargs[-num_kwarg_defaults:], node.kw_defaults)
      else:
        args += node.kwonlyargs
        default_kwargs = []
      for name, value in default_kwargs:
        args.append(KeywordArg(name, value))

    if node.kwarg:
      args.append(DoubleStarArg(ast.Name(id=node.kwarg)))

    if args:
      # lambdas can't have a multiline arglist
      allow_newlines = not isinstance(self.get_parent_node(), ast.Lambda)
      self.write_expression_list(
        args,
        allow_newlines=allow_newlines,
        need_parens=False,
        final_separator_if_multiline=False  # illegal after **kwargs
      )

  def visit_arg(self, node):
    self.write(node.arg)
    if node.annotation:
      self.write(': ')
      self.visit(node.annotation)

  def visit_keyword(self, node):
    if node.arg is None:
      # in py3, **kwargs is a keyword whose arg is None
      self.write('**')
    else:
      self.write(node.arg + '=')
    self.visit(node.value)

  def visit_alias(self, node):
    self.write(node.name)
    if node.asname is not None:
      self.write(' as %s' % node.asname)



# -*- coding: utf-8 -*-

from typing import Union

import jax.nn
import jax.numpy as jnp

import brainpy.math as bm
from brainpy.initialize import ZeroInit, OneInit, Initializer
from brainpy.nn.base import Node


__all__ = [
  'BatchNorm',
  'BatchNorm1d',
  'BatchNorm2d',
  'BatchNorm3d',
  'GroupNorm',
  'LayerNorm',
  'InstanceNorm',
]


class BatchNorm(Node):
  """Batch Normalization node.
  This layer aims to reduce the internal covariant shift of data. It
  normalizes a batch of data by fixing the mean and variance of inputs
  on each feature (channel). Most commonly, the first axis of the data
  is the batch, and the last is the channel. However, users can specify
  the axes to be normalized.

  adapted from jax.example_libraries.stax.BatchNorm
  https://jax.readthedocs.io/en/latest/_modules/jax/example_libraries/stax.html#BatchNorm

  Parameters
  ----------
  axis: int, tuple, list
    axes where the data will be normalized. The feature (channel) axis should be excluded.
  epsilon: float
    a value added to the denominator for numerical stability. Default: 1e-5
  use_bias: bool
    whether to translate data in refactoring. Default: True
  use_scale: bool
    whether to scale data in refactoring. Default: True
  beta_init: brainpy.init.Initializer
    an initializer generating the original translation matrix
  gamma_init: brainpy.init.Initializer
    an initializer generating the original scaling matrix
  """

  def __init__(self,
               axis: Union[int, tuple, list],
               epsilon: float = 1e-5,
               use_bias: bool = True,
               use_scale: bool = True,
               beta_init: Initializer = ZeroInit(),
               gamma_init: Initializer = OneInit(),
               **kwargs):
    super(BatchNorm, self).__init__(**kwargs)
    self.epsilon = epsilon
    self.bias = use_bias
    self.scale = use_scale
    self.beta_init = beta_init if use_bias else ()
    self.gamma_init = gamma_init if use_scale else ()
    self.axis = (axis,) if jnp.isscalar(axis) else axis

  def _check_input_dim(self):
    pass

  def init_ff_conn(self):
    self._check_input_dim()

    input_shape = tuple(d for i, d in enumerate(self.feedforward_shapes) if i not in self.axis)
    self.beta = bm.TrainVar(self.beta_init(input_shape)) if self.bias else None
    self.gamma = bm.TrainVar(self.gamma_init(input_shape)) if self.scale else None
    self.set_output_shape(self.feedforward_shapes)

  def forward(self, ff, **shared_kwargs):
    ed = tuple(None if i in self.axis else slice(None) for i in range(jnp.ndim(ff)))
    output = jax.nn.normalize(bm.as_device_array(ff), self.axis, epsilon=self.epsilon)
    if self.bias and self.scale: return self.gamma[ed] * output + self.beta[ed]
    if self.bias: return output + self.beta[ed]
    if self.scale: return self.gamma[ed] * output
    return output


class BatchNorm1d(BatchNorm):
  """1-D batch normalization.
  The data should be of `(b, l, c)`, where `b` is the batch dimension,
  `l` is the layer dimension, and `c` is the channel dimension, or of
  '(b, c)'.

  Parameters
  ----------
  axis: int, tuple, list
    axes where the data will be normalized. The feature (channel) axis should be excluded.
  epsilon: float
    a value added to the denominator for numerical stability. Default: 1e-5
  use_bias: bool
    whether to translate data in refactoring. Default: True
  use_scale: bool
    whether to scale data in refactoring. Default: True
  beta_init: brainpy.init.Initializer
    an initializer generating the original translation matrix
  gamma_init: brainpy.init.Initializer
    an initializer generating the original scaling matrix
  """
  def __init__(self, axis=(0, 1), **kwargs):
    super(BatchNorm1d, self).__init__(axis=axis, **kwargs)

  def _check_input_dim(self):
    ndim = len(self.feedforward_shapes)
    if ndim != 2 and ndim != 3:
      raise ValueError(
        "expected 2D or 3D input (got {}D input)".format(ndim)
      )
    if ndim == 2 and len(self.axis) == 2:
      self.axis = (0,)


class BatchNorm2d(BatchNorm):
  """2-D batch normalization.
    The data should be of `(b, h, w, c)`, where `b` is the batch dimension,
    `h` is the height dimension, `w` is the width dimension, and `c` is the
    channel dimension.

    Parameters
    ----------
    axis: int, tuple, list
      axes where the data will be normalized. The feature (channel) axis should be excluded.
    epsilon: float
      a value added to the denominator for numerical stability. Default: 1e-5
    use_bias: bool
      whether to translate data in refactoring. Default: True
    use_scale: bool
      whether to scale data in refactoring. Default: True
    beta_init: brainpy.init.Initializer
      an initializer generating the original translation matrix
    gamma_init: brainpy.init.Initializer
      an initializer generating the original scaling matrix
    """
  def __init__(self, axis=(0, 1, 2), **kwargs):
    super(BatchNorm2d, self).__init__(axis=axis, **kwargs)

  def _check_input_dim(self):
    ndim = len(self.feedforward_shapes)
    if ndim != 4:
      raise ValueError(
        "expected 4D input (got {}D input)".format(ndim)
      )


class BatchNorm3d(BatchNorm):
  """3-D batch normalization.
    The data should be of `(b, h, w, d, c)`, where `b` is the batch dimension,
    `h` is the height dimension, `w` is the width dimension, `d` is the depth
    dimension, and `c` is the channel dimension.

    Parameters
    ----------
    axis: int, tuple, list
      axes where the data will be normalized. The feature (channel) axis should be excluded.
    epsilon: float
      a value added to the denominator for numerical stability. Default: 1e-5
    use_bias: bool
      whether to translate data in refactoring. Default: True
    use_scale: bool
      whether to scale data in refactoring. Default: True
    beta_init: brainpy.init.Initializer
      an initializer generating the original translation matrix
    gamma_init: brainpy.init.Initializer
      an initializer generating the original scaling matrix
      """
  def __init__(self, axis=(0, 1, 2, 3), **kwargs):
    super(BatchNorm3d, self).__init__(axis=axis, **kwargs)

  def _check_input_dim(self):
    ndim = len(self.feedforward_shapes)
    if ndim != 5:
      raise ValueError(
        "expected 5D input (got {}D input)".format(ndim)
      )


class LayerNorm(Node):
  """Layer normalization (https://arxiv.org/abs/1607.06450).

  This layer normalizes data on each example, independently of the batch. More
  specifically, it normalizes data of shape (b, d1, d2, ..., c) on the axes of
  the data dimensions and the channel (d1, d2, ..., c). Different from batch
  normalization, gamma and beta are assigned to each position (elementwise
  operation) instead of the whole channel. If users want to assign a single
  gamma and beta to a whole example/whole channel, please use GroupNorm/
  InstanceNorm.

  Parameters
  ----------
  epsilon: float
    a value added to the denominator for numerical stability. Default: 1e-5
  use_bias: bool
    whether to translate data in refactoring. Default: True
  use_scale: bool
    whether to scale data in refactoring. Default: True
  beta_init: brainpy.init.Initializer
    an initializer generating the original translation matrix
  gamma_init: brainpy.init.Initializer
    an initializer generating the original scaling matrix
  axis: int, tuple, list
    axes where the data will be normalized. The batch axis should be excluded.
  """
  def __init__(self,
               epsilon: float = 1e-5,
               use_bias: bool = True,
               use_scale: bool = True,
               beta_init: Initializer = ZeroInit(),
               gamma_init: Initializer = OneInit(),
               axis: Union[int, tuple] = None,
               **kwargs):
    super(LayerNorm, self).__init__(**kwargs)
    self.epsilon = epsilon
    self.bias = use_bias
    self.scale = use_scale
    self.beta_init = beta_init if use_bias else ()
    self.gamma_init = gamma_init if use_scale else ()
    self.axis = (axis,) if jnp.isscalar(axis) else axis

  def default_axis(self):
    # default: the first axis (batch dim) is excluded
    return tuple(i for i in range(1, len(self.feedforward_shapes)))

  def init_ff_conn(self):
    if self.axis is None:
      self.axis = self.default_axis()
    # todo: what if elementwise_affine = False?
    input_shape = tuple(d for i, d in enumerate(self.feedforward_shapes) if i in self.axis)
    self.beta = bm.TrainVar(self.beta_init(input_shape)) if self.bias else None
    self.gamma = bm.TrainVar(self.gamma_init(input_shape)) if self.scale else None
    self.set_output_shape(self.feedforward_shapes)

  def forward(self, ff, **shared_kwargs):
    ed = tuple(None if i not in self.axis else slice(None) for i in range(jnp.ndim(ff)))
    output = jax.nn.normalize(bm.as_device_array(ff), self.axis, epsilon=self.epsilon)
    if self.bias and self.scale: return self.gamma[ed] * output + self.beta[ed]
    if self.bias: return output + self.beta[ed]
    if self.scale: return self.gamma[ed] * output
    return output


class GroupNorm(Node):
  """Group normalization layer.

  This layer divides channels into groups and normalizes the features within each
  group. Its computation is also independent of the batch size. The feature size
  must be multiple of the group size.

  The shape of the data should be (b, d1, d2, ..., c), where `d` denotes the batch
  size and `c` denotes the feature (channel) size. The `d` and `c` axis should be
  excluded in parameter `axis`.

  Parameters
  ----------
  num_groups: int
    the number of groups. It should be a factor of the number of features.
  group_size: int
    the group size. It should equal to int(num_features / num_groups).
    Either `num_groups` or `group_size` should be specified.
  epsilon: float
    a value added to the denominator for numerical stability. Default: 1e-5
  use_bias: bool
    whether to translate data in refactoring. Default: True
  use_scale: bool
    whether to scale data in refactoring. Default: True
  beta_init: brainpy.init.Initializer
    an initializer generating the original translation matrix
  gamma_init: brainpy.init.Initializer
    an initializer generating the original scaling matrix
  axis: int, tuple, list
    axes where the data will be normalized. Besides the batch axis, the channel
    axis should be also excluded, since it will be automatically added to `axis`.
  """
  def __init__(self,
               num_groups: int = None,
               group_size: int = None,
               epsilon: float = 1e-5,
               use_bias: bool = True,
               use_scale: bool = True,
               beta_init: Initializer = ZeroInit(),
               gamma_init: Initializer = OneInit(),
               axis: Union[int, tuple] = None,
               **kwargs):
    super(GroupNorm, self).__init__(**kwargs)
    self.num_groups = num_groups
    self.group_size = group_size
    self.epsilon = epsilon
    self.bias = use_bias
    self.scale = use_scale
    self.beta_init = beta_init if use_bias else ()
    self.gamma_init = gamma_init if use_scale else ()
    self.norm_axis = (axis,) if jnp.isscalar(axis) else axis

  def init_ff_conn(self):
    num_channels = self.feedforward_shapes[-1]
    self.ndim = len(self.feedforward_shapes)

    # compute num_groups and group_size
    if ((self.num_groups is None and self.group_size is None) or
        (self.num_groups is not None and self.group_size is not None)):
      raise ValueError('Either `num_groups` or `group_size` should be specified. '
                       'Once one is specified, the other will be automatically '
                       'computed.')

    if self.num_groups is None:
      assert self.group_size > 0, '`group_size` should be a positive integer.'
      if num_channels % self.group_size != 0:
        raise ValueError('The number of channels ({}) is not multiple of the '
                         'group size ({}).'.format(num_channels, self.group_size))
      else:
        self.num_groups = num_channels // self.group_size
    else:  # self.num_groups is not None:
      assert self.num_groups > 0, '`num_groups` should be a positive integer.'
      if num_channels % self.num_groups != 0:
        raise ValueError('The number of channels ({}) is not multiple of the '
                         'number of groups ({}).'.format(num_channels, self.num_groups))
      else:
        self.group_size = num_channels // self.num_groups

    # axes for normalization
    if self.norm_axis is None:
      # default: the first axis (batch dim) and the second-last axis (num_group dim) are excluded
      self.norm_axis = tuple(i for i in range(1, len(self.feedforward_shapes) - 1)) + (self.ndim,)

    group_shape = self.feedforward_shapes[:-1] + (self.num_groups, self.group_size)
    input_shape = tuple(d for i, d in enumerate(group_shape) if i in self.norm_axis)
    self.beta = bm.TrainVar(self.beta_init(input_shape)) if self.bias else None
    self.gamma = bm.TrainVar(self.gamma_init(input_shape)) if self.scale else None
    self.set_output_shape(self.feedforward_shapes)

  def forward(self, ff, **shared_kwargs):
    group_shape = ff.shape[:-1] + (self.num_groups, self.group_size)
    ff_reshape = ff.reshape(group_shape)
    ed = tuple(None if i not in self.norm_axis else slice(None) for i in range(jnp.ndim(ff_reshape)))
    output = jax.nn.normalize(bm.as_device_array(ff_reshape), self.norm_axis, epsilon=self.epsilon)
    if self.bias and self.scale:
      output = self.gamma[ed] * output + self.beta[ed]
    elif self.bias:
      output = output + self.beta[ed]
    elif self.scale:
      output = self.gamma[ed] * output
    return output.reshape(ff.shape)


class InstanceNorm(GroupNorm):
  """Instance normalization layer.

  This layer normalizes the data within each feature. It can be regarded as
  a group normalization layer in which `group_size` equals to 1.

  Parameters
  ----------
  epsilon: float
    a value added to the denominator for numerical stability. Default: 1e-5
  use_bias: bool
    whether to translate data in refactoring. Default: True
  use_scale: bool
    whether to scale data in refactoring. Default: True
  beta_init: brainpy.init.Initializer
    an initializer generating the original translation matrix
  gamma_init: brainpy.init.Initializer
    an initializer generating the original scaling matrix
  axis: int, tuple, list
    axes where the data will be normalized. The batch and channel axes
    should be excluded.
  """
  def __init__(self,
               epsilon: float = 1e-5,
               use_bias: bool = True,
               use_scale: bool = True,
               beta_init: Initializer = ZeroInit(),
               gamma_init: Initializer = OneInit(),
               axis: Union[int, tuple] = None,
               **kwargs):
    super(InstanceNorm, self).__init__(group_size=1, epsilon=epsilon, use_bias=use_bias,
                                       use_scale=use_scale, beta_init=beta_init,
                                       gamma_init=gamma_init, axis=axis, **kwargs)



# -*- coding: utf-8 -*-

import numpy as np


def add_arrow(line, position=None, direction='right', size=15, color=None):
  """
  add an arrow to a line.

  line:       Line2D object
  position:   x-position of the arrow. If None, mean of xdata is taken
  direction:  'left' or 'right'
  size:       size of the arrow in fontsize points
  color:      if None, line color is taken.
  """
  if color is None:
    color = line.get_color()

  xdata = line.get_xdata()
  ydata = line.get_ydata()

  if position is None:
    position = xdata.mean()
  # find closest index
  start_ind = np.argmin(np.absolute(xdata - position))
  if direction == 'right':
    end_ind = start_ind + 1
  else:
    end_ind = start_ind - 1

  line.axes.annotate(text='',
                     xytext=(xdata[start_ind], ydata[start_ind]),
                     xy=(xdata[end_ind], ydata[end_ind]),
                     arrowprops=dict(arrowstyle="->", color=color),
                     size=size)



# -*- coding: utf-8 -*-

import time
from typing import Union, Dict, Sequence, Callable

import jax.numpy as jnp
import numpy as np
import tqdm.auto
from jax.experimental.host_callback import id_tap

from brainpy import math as bm
from brainpy.base.collector import Collector, TensorCollector
from brainpy.errors import RunningError, MonitorError
from brainpy.integrators.base import Integrator
from brainpy.running.runner import Runner


__all__ = [
  'IntegratorRunner',
]


class IntegratorRunner(Runner):
  """Structural runner for numerical integrators in brainpy.

  Examples
  --------

  Example to run an ODE integrator,

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>> a=0.7; b=0.8; tau=12.5
    >>> dV = lambda V, t, w, I: V - V * V * V / 3 - w + I
    >>> dw = lambda w, t, V, a, b: (V + a - b * w) / tau
    >>> integral = bp.odeint(bp.JointEq([dV, dw]), method='exp_auto')
    >>>
    >>> runner = bp.integrators.IntegratorRunner(
    >>>          integral,  # the simulation target
    >>>          monitors=['V', 'w'],  # the variables to monitor
    >>>          inits={'V': bm.random.rand(10),
    >>>                 'w': bm.random.normal(size=10)},  # the initial values
    >>>          args={'a': 1., 'b': 1.},  # update arguments
    >>>          dyn_args={'I': bp.inputs.ramp_input(0, 4, 200)},  # each time each current input
    >>> )
    >>> runner.run(100.)
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, plot_ids=[0, 1, 4], show=True)

  Example to run an SDE intragetor,

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>> # stochastic Lorenz system
    >>> sigma=10; beta=8 / 3; rho=28
    >>> g = lambda x, y, z, t, p: (p * x, p * y, p * z)
    >>> f = lambda x, y, z, t, p: [sigma * (y - x), x * (rho - z) - y, x * y - beta * z]
    >>> lorenz = bp.sdeint(f, g, method='milstein')
    >>>
    >>> runner = bp.integrators.IntegratorRunner(
    >>>   lorenz,
    >>>   monitors=['x', 'y', 'z'],
    >>>   inits=[1., 1., 1.], # initialize all variable to 1.
    >>>   args={'p': 0.1},
    >>>   dt=0.01
    >>> )
    >>> runner.run(100.)
    >>>
    >>> import matplotlib.pyplot as plt
    >>> fig = plt.figure()
    >>> ax = fig.gca(projection='3d')
    >>> plt.plot(runner.mon.x.squeeze(), runner.mon.y.squeeze(), runner.mon.z.squeeze())
    >>> ax.set_xlabel('x')
    >>> ax.set_xlabel('y')
    >>> ax.set_xlabel('z')
    >>> plt.show()

  Example to run an DDE integrator,

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import brainpy.math as bm
    >>> import matplotlib.pyplot as plt
    >>>
    >>> # Mackey-Glass equation
    >>> dt = 0.01; beta=2.; gamma=1.; tau=2.; n=9.65
    >>> mg_eq = lambda x, t, xdelay: (beta * xdelay(t - tau) / (1 + xdelay(t - tau) ** n)
    >>>                               - gamma * x)
    >>> xdelay = bm.TimeDelay(bm.asarray([1.2]), delay_len=tau, dt=dt, before_t0=lambda t: 1.2)
    >>> integral = bp.ddeint(mg_eq, method='rk4', state_delays={'x': xdelay})
    >>> runner = bp.integrators.IntegratorRunner(
    >>>       integral,
    >>>       monitors=['x', ],
    >>>       fun_monitors={'x(tau)': (lambda t, _: xdelay(t - tau))},
    >>>       inits=[1.2],  # initialize all variable to 1.
    >>>       args={'xdelay': xdelay}, dt=dt,
    >>> )
    >>> runner.run(100.)
    >>> plt.plot(runner.mon['x'].flatten(), runner.mon['x(tau)'].flatten())
    >>> plt.show()

  """

  def __init__(
      self,
      target: Integrator,

      # IntegratorRunner specific arguments
      inits: Union[Sequence, Dict] = None,
      args: Dict = None,
      dyn_args: Dict[str, Union[bm.ndarray, jnp.ndarray]] = None,
      dt: Union[float, int] = None,

      # regular/common arguments
      fun_monitors: Dict[str, Callable] = None,
      monitors: Sequence[str] = None,
      dyn_vars: Dict[str, bm.Variable] = None,
      jit: bool = True,
      numpy_mon_after_run: bool = True,
      progress_bar: bool = True
  ):
    """Initialization of structural runner for integrators.

    Parameters
    ----------
    target: Integrator
    monitors: sequence of str
    fun_monitors: dict
    inits: sequence, dict
      The initial value of variables. With this parameter,
      you can easily control the number of variables to simulate.
      For example, if one of the variable has the shape of 10,
      then all variables will be an instance of :py:class:`brainpy.math.Variable`
      with the shape of :math:`(10,)`.
    args: dict
      The equation arguments to update.
      Note that if one of the arguments are heterogeneous (i.e., a tensor),
      it means we should run multiple trials. However, you can set the number
      of the elements in the variables so that each pair of variables can
      corresponds to one set of arguments.
    dyn_args: dict
      The dynamically changed arguments. This means this argument can control
      the argument dynamically changed. For example, if you want to inject a
      time varied currents into the HH neuron model, you can pack the currents
      into this ``dyn_args`` argument.
    dt: float, int
    dyn_vars: dict
    jit: bool
    progress_bar: bool
    numpy_mon_after_run: bool
    """
    super(IntegratorRunner, self).__init__(target=target,
                                           monitors=monitors,
                                           fun_monitors=fun_monitors,
                                           jit=jit,
                                           progress_bar=progress_bar,
                                           dyn_vars=dyn_vars,
                                           numpy_mon_after_run=numpy_mon_after_run)

    # parameters
    dt = bm.get_dt() if dt is None else dt
    if not isinstance(dt, (int, float)):
      raise RunningError(f'"dt" must be scalar, but got {dt}')
    self.dt = dt

    # target
    if not isinstance(self.target, Integrator):
      raise RunningError(f'"target" must be an instance of {Integrator.__name__}, '
                         f'but we got {type(target)}: {target}')

    # arguments of the integral function
    self._static_args = Collector()
    if args is not None:
      assert isinstance(args, dict), (f'"args" must be a dict, but '
                                      f'we got {type(args)}: {args}')
      self._static_args.update(args)
    self._dyn_args = Collector()
    if dyn_args is not None:
      assert isinstance(dyn_args, dict), (f'"dyn_args" must be a dict, but we get '
                                          f'{type(dyn_args)}: {dyn_args}')
      sizes = np.unique([len(v) for v in dyn_args.values()])
      num_size = len(sizes)
      if num_size != 1:
        raise RunningError(f'All values in "dyn_args" should have the same length. '
                           f'But we got {num_size}: {sizes}')
      self._dyn_args.update(dyn_args)

    # monitors
    for k in self.mon.item_names:
      if k not in self.target.variables:
        raise MonitorError(f'Variable "{k}" to monitor is not defined '
                           f'in the integrator {self.target}.')

    # start simulation time
    self._start_t = None

    # dynamical changed variables
    self.dyn_vars.update(self.target.vars().unique())

    # Variables
    if inits is not None:
      if isinstance(inits, (list, tuple)):
        assert len(self.target.variables) == len(inits)
        inits = {k: inits[i] for i, k in enumerate(self.target.variables)}
      assert isinstance(inits, dict), f'"inits" must be a dict, but we got {type(inits)}'
      sizes = np.unique([np.size(v) for v in list(inits.values())])
      max_size = np.max(sizes)
    else:
      max_size = 1
      inits = dict()
    self.variables = TensorCollector({v: bm.Variable(bm.zeros(max_size))
                                      for v in self.target.variables})
    for k in inits.keys():
      self.variables[k][:] = inits[k]
    self.dyn_vars.update(self.variables)
    if len(self._dyn_args) > 0:
      self.idx = bm.Variable(bm.zeros(1, dtype=bm.int_))
      self.dyn_vars['_idx'] = self.idx

    # build the update step
    if jit:
      _loop_func = bm.make_loop(
        self._step,
        dyn_vars=self.dyn_vars,
        out_vars={k: self.variables[k] for k in self.mon.item_names},
        has_return=True
      )
    else:
      def _loop_func(times):
        out_vars = {k: [] for k in self.mon.item_names}
        returns = {k: [] for k in self.fun_monitors.keys()}
        for i in range(len(times)):
          _t = times[i]
          _dt = self.dt
          # function monitor
          for k, v in self.fun_monitors.items():
            returns[k].append(v(_t, _dt))
          # step call
          self._step(_t)
          # variable monitors
          for k in self.mon.item_names:
            out_vars[k].append(bm.as_device_array(self.variables[k]))
        out_vars = {k: bm.asarray(out_vars[k]) for k in self.mon.item_names}
        return out_vars, returns
    self.step_func = _loop_func

  def _post(self, times, returns: dict):  # monitor
    self.mon.ts = times + self.dt
    for key in returns.keys():
      self.mon.item_contents[key] = bm.asarray(returns[key])

  def _step(self, t):
    # arguments
    kwargs = dict()
    kwargs.update(self.variables)
    kwargs.update({'t': t, 'dt': self.dt})
    kwargs.update(self._static_args)
    if len(self._dyn_args) > 0:
      kwargs.update({k: v[self.idx.value] for k, v in self._dyn_args.items()})
      self.idx += 1
    # return of function monitors
    returns = dict()
    for key, func in self.fun_monitors.items():
      returns[key] = func(t, self.dt)
    # call integrator function
    update_values = self.target(**kwargs)
    if len(self.target.variables) == 1:
      self.variables[self.target.variables[0]].update(update_values)
    else:
      for i, v in enumerate(self.target.variables):
        self.variables[v].update(update_values[i])
    if self.progress_bar:
      id_tap(lambda *args: self._pbar.update(), ())
    return returns

  def run(self, duration, start_t=None):
    self.__call__(duration, start_t)

  def __call__(self, duration, start_t=None):
    """The running function.

    Parameters
    ----------
    duration : float, int, tuple, list
      The running duration.
    start_t : float, optional

    Returns
    -------
    running_time : float
      The total running time.
    """
    if len(self._dyn_args) > 0:
      self.dyn_vars['_idx'][0] = 0

    # time step
    if start_t is None:
      if self._start_t is None:
        start_t = 0.
      else:
        start_t = float(self._start_t)
    end_t = float(start_t + duration)
    # times
    times = np.arange(start_t, end_t, self.dt)

    # running
    if self.progress_bar:
      self._pbar = tqdm.auto.tqdm(total=times.size)
      self._pbar.set_description(f"Running a duration of {round(float(duration), 3)} ({times.size} steps)",
                                 refresh=True)
    t0 = time.time()
    hists, returns = self.step_func(times)
    running_time = time.time() - t0
    if self.progress_bar:
      self._pbar.close()
    # post-running
    hists.update(returns)
    self._post(times, hists)
    self._start_t = end_t
    if self.numpy_mon_after_run:
      self.mon.numpy()
    return running_time



# -*- coding: utf-8 -*-


from typing import Dict

from brainpy import math as bm
from brainpy.errors import DiffEqError, CodeError
from brainpy.integrators import constants, utils
from brainpy.integrators.base import Integrator
from brainpy.integrators.constants import DT
from brainpy.tools.checking import check_dict_data

__all__ = [
  'ODEIntegrator',
]


def f_names(f):
  func_name = constants.unique_name('ode')
  if f.__name__.isidentifier():
    func_name += '_' + f.__name__
  return func_name


class ODEIntegrator(Integrator):
  """Numerical Integrator for Ordinary Differential Equations (ODEs).

  Parameters
  ----------
  f : callable
    The derivative function.
  var_type: str
    The type for each variable.
  dt: float, int
    The numerical precision.
  name: str
    The integrator name.
  """

  def __init__(
      self,
      f,
      var_type=None,
      dt=None,
      name=None,
      show_code=False,
      state_delays: Dict[str, bm.AbstractDelay] = None,
      neutral_delays: Dict[str, bm.NeutralDelay] = None
  ):

    dt = bm.get_dt() if dt is None else dt
    parses = utils.get_args(f)
    variables = parses[0]  # variable names, (before 't')
    parameters = parses[1]  # parameter names, (after 't')
    arguments = parses[2]  # function arguments

    for p in tuple(variables) + tuple(parameters):
      if p == DT:
        raise CodeError(f'{DT} is a system keyword denotes the '
                        f'precision of numerical integration. '
                        f'It cannot be used as a variable or parameter, '
                        f'please change an another name.')

    # super initialization
    super(ODEIntegrator, self).__init__(name=name,
                                        variables=variables,
                                        parameters=parameters,
                                        arguments=arguments,
                                        dt=dt,
                                        state_delays=state_delays)

    # others
    self.show_code = show_code
    self.var_type = var_type  # variable type

    # derivative function
    self.derivative = {constants.F: f}
    self.f = f

    # code scope
    self.code_scope = {constants.F: f}

    # code lines
    self.func_name = f_names(f)
    self.code_lines = [f'def {self.func_name}({", ".join(self.arguments)}):']

    # neutral delays
    self._neutral_delays = dict()
    if neutral_delays is not None:
      check_dict_data(neutral_delays, key_type=str, val_type=bm.NeutralDelay)
      for key, delay in neutral_delays.items():
        if key not in self.variables:
          raise DiffEqError(f'"{key}" is not defined in the variables: {self.variables}')
        self._neutral_delays[key] = delay
    self.register_implicit_nodes(self._neutral_delays)

  @property
  def neutral_delays(self):
    """neutral delays."""
    return self._neutral_delays

  @neutral_delays.setter
  def neutral_delays(self, value):
    raise ValueError('Cannot set "neutral_delays" by users.')

  def __call__(self, *args, **kwargs):
    assert self.integral is not None, 'Please build the integrator first.'

    # check arguments
    for i, arg in enumerate(args):
      kwargs[self.arg_names[i]] = arg

    # integral
    new_vars = self.integral(**kwargs)
    if len(self.variables) == 1:
      dict_vars = {self.variables[0]: new_vars}
    else:
      dict_vars = {k: new_vars[i] for i, k in enumerate(self.variables)}

    dt = kwargs.pop(DT, self.dt)
    # update neutral delay variables
    if len(self.neutral_delays):
      kwargs.update(dict_vars)
      new_devs = self.f(**kwargs)
      if len(self.variables) == 1:
        new_devs = {self.variables[0]: new_devs}
      else:
        new_devs = {k: new_devs[i] for i, k in enumerate(self.variables)}
      for key, delay in self.neutral_delays.items():
        if isinstance(delay, bm.LengthDelay):
          delay.update(new_devs[key])
        elif isinstance(delay, bm.TimeDelay):
          delay.update(kwargs['t'] + dt, new_devs[key])
        else:
          raise ValueError('Unknown delay variable. We only supports '
                           'brainpy.math.LengthDelay, brainpy.math.TimeDelay, '
                           'brainpy.math.NeutralDelay. '
                           f'While we got {delay}')

    # update state delay variables
    for key, delay in self.state_delays.items():
      if isinstance(delay, bm.LengthDelay):
        delay.update(dict_vars[key])
      elif isinstance(delay, bm.TimeDelay):
        delay.update(kwargs['t'] + dt, dict_vars[key])
      else:
        raise ValueError('Unknown delay variable. We only supports '
                         'brainpy.math.LengthDelay, brainpy.math.TimeDelay, '
                         'brainpy.math.NeutralDelay. '
                         f'While we got {delay}')

    return new_vars



# -*- coding: utf-8 -*-

import jax.numpy as jnp
import numpy as np

from brainpy import tools
from brainpy.errors import ConnectorError
from brainpy.math.jaxarray import JaxArray
from .base import *

__all__ = [
  'MatConn',
  'IJConn',
  'SparseMatConn'
]


class MatConn(TwoEndConnector):
  """Connector built from the dense connection matrix."""

  def __init__(self, conn_mat):
    super(MatConn, self).__init__()

    assert isinstance(conn_mat, (np.ndarray, JaxArray, jnp.ndarray)) and conn_mat.ndim == 2
    self.pre_num, self.post_num = conn_mat.shape
    self.pre_size, self.post_size = (self.pre_num,), (self.post_num,)
    
    self.conn_mat = np.asarray(conn_mat).astype(MAT_DTYPE)
  
  def __call__(self, pre_size, post_size):
    assert self.pre_num == tools.size2num(pre_size)
    assert self.post_num == tools.size2num(post_size)
    return self

  def build_conn(self):
    return 'mat', self.conn_mat


class IJConn(TwoEndConnector):
  """Connector built from the ``pre_ids`` and ``post_ids`` connections."""

  def __init__(self, i, j):
    super(IJConn, self).__init__()

    assert isinstance(i, (np.ndarray, JaxArray, jnp.ndarray)) and i.ndim == 1
    assert isinstance(j, (np.ndarray, JaxArray, jnp.ndarray)) and j.ndim == 1
    assert i.size == j.size

    # initialize the class via "pre_ids" and "post_ids"
    self.pre_ids = np.asarray(i).astype(IDX_DTYPE)
    self.post_ids = np.asarray(j).astype(IDX_DTYPE)

  def __call__(self, pre_size, post_size):
    super(IJConn, self).__call__(pre_size, post_size)

    max_pre = np.max(self.pre_ids)
    max_post = np.max(self.post_ids)
    if max_pre >= self.pre_num:
      raise ConnectorError(f'pre_num ({self.pre_num}) should be greater than '
                           f'the maximum id ({max_pre}) of self.pre_ids.')
    if max_post >= self.post_num:
      raise ConnectorError(f'post_num ({self.post_num}) should be greater than '
                           f'the maximum id ({max_post}) of self.post_ids.')
    return self

  def build_conn(self):
    return 'ij', (self.pre_ids, self.post_ids)


class SparseMatConn(TwoEndConnector):
  """Connector built from the sparse connection matrix"""

  def __init__(self, csr_mat):
    super(SparseMatConn, self).__init__()

    try:
      from scipy.sparse import csr_matrix
    except (ModuleNotFoundError, ImportError):
      raise ConnectorError(f'Using SparseMatConn requires the scipy package. '
                           f'Please run "pip install scipy" to install scipy.')

    assert isinstance(csr_mat, csr_matrix)
    csr_mat.data = np.asarray(csr_mat.data).astype(MAT_DTYPE)
    self.csr_mat = csr_mat
    self.pre_num, self.post_num = csr_mat.shape

  def __call__(self, pre_size, post_size):
    try:
      assert self.pre_num == tools.size2num(pre_size)
      assert self.post_num == tools.size2num(post_size)
    except AssertionError:
      raise ConnectorError(f'(pre_size, post_size) is inconsistent with the shape of the sparse matrix.')

    super(SparseMatConn, self).__call__(pre_size, post_size)
    return self

  def build_conn(self):
    ind, indptr = self.csr_mat.indices, self.csr_mat.indptr
    return 'csr', (ind, indptr)



# -*- coding: utf-8 -*-


class BrainPyError(Exception):
  """General BrainPy error."""
  pass


class ModelBuildError(BrainPyError):
  """The error occurred during the definition of models."""
  pass


class RunningError(BrainPyError):
  """The error occurred in the running function."""
  pass


class IntegratorError(BrainPyError):
  pass


class DiffEqError(BrainPyError):
  """The differential equation definition error."""
  pass


class CodeError(BrainPyError):
  """Code definition error.
  """
  pass


class AnalyzerError(BrainPyError):
  """Error occurred in differential equation analyzer and dynamics analysis.
  """


class PackageMissingError(BrainPyError):
  """The package missing error.
  """
  pass


class BackendNotInstalled(BrainPyError):
  def __init__(self, backend):
    super(BackendNotInstalled, self).__init__(
      '"{bk}" must be installed when the user wants to use {bk} backend. \n'
      'Please install {bk} through "pip install {bk}" '
      'or "conda install {bk}".'.format(bk=backend))


class UniqueNameError(BrainPyError):
  def __init__(self, *args):
    super(UniqueNameError, self).__init__(*args)


class UnsupportedError(BrainPyError):
  pass


class NoImplementationError(BrainPyError):
  pass


class NoLongerSupportError(BrainPyError):
  pass


class ConnectorError(BrainPyError):
  pass


class MonitorError(BrainPyError):
  pass


class MathError(BrainPyError):
  """Errors occurred in ``brainpy.math`` module."""
  pass


class JaxTracerError(MathError):
  def __init__(self, variables=None):
    msg = 'There is an unexpected tracer. \n\n' \
          'In BrainPy, all the dynamically changed variables must be declared as ' \
          '"brainpy.math.Variable" and they should be provided ' \
          'into the "dyn_vars" when calling the transformation functions, ' \
          'like "jit()", "vmap()", "grad()", "make_loop()", etc. \n\n'

    if variables is None:
      pass
    elif isinstance(variables, dict):
      msg += f'We detect all the provided dynamical variables are: ' \
             f'{variables.keys()}\n\n'
    elif isinstance(variables, (list, tuple)):
      msg += 'We detect all the provided dynamical variables are: \n'
      for v in variables:
        msg += f'\t{v.dtype}[{v.shape}]\n'
      msg += '\n'
    else:
      raise ValueError

    # msg += 'While there are changed variables which are not wrapped into "dyn_vars". Please check!'
    msg = 'While there are changed variables which are not wrapped into "dyn_vars". Please check!'

    super(JaxTracerError, self).__init__(msg)


class ConcretizationTypeError(Exception):
  def __init__(self):
    super(ConcretizationTypeError, self).__init__(
      'This problem may be caused by two ways:\n'
      '1. You did not set the "static_argnames" for your jitted function. '
      'More details please see https://jax.readthedocs.io/en/latest/errors.html#jax.errors.ConcretizationTypeError\n'
      '2. The static variables which set in the "static_argnames" are not provided '
      'as keyword arguments, like "jit_fun(static_key1=v1, static_key2=v2)".'
    )




# -*- coding: utf-8 -*-

import jax.numpy as jnp
import numpy as np
from jax import vmap

import brainpy.math as bm
from brainpy import errors, math
from brainpy.analysis import stability, constants as C, utils
from brainpy.analysis.lowdim.lowdim_analyzer import *

pyplot = None

__all__ = [
  'PhasePlane1D',
  'PhasePlane2D',
]


class PhasePlane1D(Num1DAnalyzer):
  """Phase plane analyzer for 1D dynamical system.

  This class can help users fast check:

  - Vector fields
  - Fixed points

  Parameters
  ----------
  model : Any
      A model of the population, the integrator function,
      or a list/tuple of integrator functions.
  target_vars : dict
      The target/dynamical variables.
  fixed_vars : dict
      The fixed variables.
  target_pars : dict, optional
      The parameters which can be dynamical varied.
  pars_update : dict, optional
      The parameters to update.
  resolutions : float, dict

  """

  def __init__(self,
               model,
               target_vars,
               fixed_vars=None,
               target_pars=None,
               pars_update=None,
               resolutions=None,
               **kwargs):
    if (target_pars is not None) and len(target_pars) > 0:
      raise errors.AnalyzerError(f'Phase plane analysis does not support "target_pars". '
                                 f'While we detect "target_pars={target_pars}".')
    super(PhasePlane1D, self).__init__(model=model,
                                       target_vars=target_vars,
                                       fixed_vars=fixed_vars,
                                       target_pars=target_pars,
                                       pars_update=pars_update,
                                       resolutions=resolutions,
                                       **kwargs)
    # utils.output(f'I am {PhasePlane1D.__name__}.')

  def plot_vector_field(self, show=False, with_plot=True, with_return=False):
    """Plot the vector filed."""
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am creating the vector field ...')

    # Nullcline of the x variable
    y_val = self.F_fx(self.resolutions[self.x_var])
    y_val = np.asarray(y_val)

    # visualization
    if with_plot:
      label = f"d{self.x_var}dt"
      x_style = dict(color='lightcoral', alpha=.7, linewidth=4)
      pyplot.plot(np.asarray(self.resolutions[self.x_var]), y_val, **x_style, label=label)
      pyplot.axhline(0)
      pyplot.xlabel(self.x_var)
      pyplot.ylabel(label)
      pyplot.xlim(*utils.rescale(self.target_vars[self.x_var], scale=(self.lim_scale - 1.) / 2))
      pyplot.legend()
      if show: pyplot.show()
    # return
    if with_return:
      return y_val

  def plot_fixed_point(self, show=False, with_plot=True, with_return=False):
    """Plot the fixed point."""
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am searching fixed points ...')

    # fixed points and stability analysis
    fps, _, pars = self._get_fixed_points(self.resolutions[self.x_var])
    container = {a: [] for a in stability.get_1d_stability_types()}
    for i in range(len(fps)):
      x = fps[i]
      dfdx = self.F_dfxdx(x)
      fp_type = stability.stability_analysis(dfdx)
      utils.output(f"Fixed point #{i + 1} at {self.x_var}={x} is a {fp_type}.")
      container[fp_type].append(x)

    # visualization
    if with_plot:
      for fp_type, points in container.items():
        if len(points):
          plot_style = stability.plot_scheme[fp_type]
          pyplot.plot(points, [0] * len(points), '.', markersize=20, **plot_style, label=fp_type)
      pyplot.legend()
      if show:
        pyplot.show()

    # return
    if with_return:
      return fps


class PhasePlane2D(Num2DAnalyzer):
  """Phase plane analyzer for 2D dynamical system.

  Parameters
  ----------
  model : Any
    A model of the population, the integrator function,
    or a list/tuple of integrator functions.
  target_vars : dict
    The target/dynamical variables.
  fixed_vars : dict
    The fixed variables.
  target_pars : dict, optional
    The parameters which can be dynamical varied.
  pars_update : dict, optional
    The parameters to update.
  resolutions : float, dict
  """

  def __init__(self,
               model,
               target_vars,
               fixed_vars=None,
               target_pars=None,
               pars_update=None,
               resolutions=None,
               **kwargs):
    if (target_pars is not None) and len(target_pars) > 0:
      raise errors.AnalyzerError(f'Phase plane analysis does not support "target_pars". '
                                 f'While we detect "target_pars={target_pars}".')
    super(PhasePlane2D, self).__init__(model=model,
                                       target_vars=target_vars,
                                       fixed_vars=fixed_vars,
                                       target_pars=target_pars,
                                       pars_update=pars_update,
                                       resolutions=resolutions,
                                       **kwargs)

  @property
  def F_vmap_brentq_fy(self):
    if C.F_vmap_brentq_fy not in self.analyzed_results:
      f_opt = bm.jit(vmap(utils.jax_brentq(self.F_fy)))
      self.analyzed_results[C.F_vmap_brentq_fy] = f_opt
    return self.analyzed_results[C.F_vmap_brentq_fy]

  def plot_vector_field(self, with_plot=True, with_return=False,
                        plot_method='streamplot', plot_style=None, show=False):
    """Plot the vector field.

    Parameters
    ----------
    with_plot: bool
    with_return : bool
    show : bool
    plot_method : str
        The method to plot the vector filed. It can be "streamplot" or "quiver".
    plot_style : dict, optional
        The style for vector filed plotting.

        - For ``plot_method="streamplot"``, it can set the keywords like "density",
          "linewidth", "color", "arrowsize". More settings please check
          https://matplotlib.org/api/_as_gen/matplotlib.pyplot.streamplot.html.
        - For ``plot_method="quiver"``, it can set the keywords like "color",
          "units", "angles", "scale". More settings please check
          https://matplotlib.org/api/_as_gen/matplotlib.pyplot.quiver.html.
    """
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am creating the vector field ...')

    # get vector fields
    xs = self.resolutions[self.x_var]
    ys = self.resolutions[self.y_var]
    X, Y = bm.meshgrid(xs, ys)
    dx = self.F_fx(X, Y)
    dy = self.F_fy(X, Y)
    X, Y = np.asarray(X), np.asarray(Y)
    dx, dy = np.asarray(dx), np.asarray(dy)

    if with_plot:  # plot vector fields
      if plot_method == 'quiver':
        if plot_style is None:
          plot_style = dict(units='xy')
        if (not np.isnan(dx).any()) and (not np.isnan(dy).any()):
          speed = np.sqrt(dx ** 2 + dy ** 2)
          dx = dx / speed
          dy = dy / speed
        pyplot.quiver(X, Y, dx, dy, **plot_style)
      elif plot_method == 'streamplot':
        if plot_style is None:
          plot_style = dict(arrowsize=1.2, density=1, color='thistle')
        linewidth = plot_style.get('linewidth', None)
        if linewidth is None:
          if (not np.isnan(dx).any()) and (not np.isnan(dy).any()):
            min_width, max_width = 0.5, 5.5
            speed = np.nan_to_num(np.sqrt(dx ** 2 + dy ** 2))
            linewidth = min_width + max_width * (speed / speed.max())
        pyplot.streamplot(X, Y, dx, dy, linewidth=linewidth, **plot_style)
      else:
        raise errors.AnalyzerError(f'Unknown plot_method "{plot_method}", '
                                   f'only supports "quiver" and "streamplot".')

      pyplot.xlabel(self.x_var)
      pyplot.ylabel(self.y_var)
      if show:
        pyplot.show()

    if with_return:  # return vector fields
      return dx, dy

  def plot_nullcline(self, with_plot=True, with_return=False,
                     y_style=None, x_style=None, show=False,
                     coords=None, tol_nullcline=1e-7):
    """Plot the nullcline."""
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am computing fx-nullcline ...')

    if coords is None:
      coords = dict()
    x_coord = coords.get(self.x_var, None)
    y_coord = coords.get(self.y_var, None)

    # Nullcline of the x variable
    xy_values_in_fx, = self._get_fx_nullcline_points(coords=x_coord, tol=tol_nullcline)
    x_values_in_fx = np.asarray(xy_values_in_fx[:, 0])
    y_values_in_fx = np.asarray(xy_values_in_fx[:, 1])

    if with_plot:
      if x_style is None:
        x_style = dict(color='cornflowerblue', alpha=.7, )
      fmt = x_style.pop('fmt', '.')
      pyplot.plot(x_values_in_fx, y_values_in_fx, fmt, **x_style, label=f"{self.x_var} nullcline")

    # Nullcline of the y variable
    utils.output('I am computing fy-nullcline ...')
    xy_values_in_fy, = self._get_fy_nullcline_points(coords=y_coord, tol=tol_nullcline)
    x_values_in_fy = np.asarray(xy_values_in_fy[:, 0])
    y_values_in_fy = np.asarray(xy_values_in_fy[:, 1])

    if with_plot:
      if y_style is None:
        y_style = dict(color='lightcoral', alpha=.7, )
      fmt = y_style.pop('fmt', '.')
      pyplot.plot(x_values_in_fy, y_values_in_fy, fmt, **y_style, label=f"{self.y_var} nullcline")

    if with_plot:
      pyplot.xlabel(self.x_var)
      pyplot.ylabel(self.y_var)
      scale = (self.lim_scale - 1.) / 2
      pyplot.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
      pyplot.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
      pyplot.legend()
      if show:
        pyplot.show()

    if with_return:
      return {self.x_var: (x_values_in_fx, y_values_in_fx),
              self.y_var: (x_values_in_fy, y_values_in_fy)}

  def plot_fixed_point(self, with_plot=True, with_return=False, show=False,
                       tol_unique=1e-2, tol_aux=1e-8, tol_opt_screen=None,
                       select_candidates='fx-nullcline', num_rank=100, ):
    """Plot the fixed point and analyze its stability.
    """
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am searching fixed points ...')

    if self._can_convert_to_one_eq():
      if self.convert_type() == C.x_by_y:
        candidates = self.resolutions[self.y_var].value
      else:
        candidates = self.resolutions[self.x_var].value
    else:
      if select_candidates == 'fx-nullcline':
        candidates = [self.analyzed_results[key][0] for key in self.analyzed_results.keys()
                      if key.startswith(C.fx_nullcline_points)]
        if len(candidates) == 0:
          raise errors.AnalyzerError(f'No nullcline points are found, please call '
                                     f'".{self.plot_nullcline.__name__}()" first.')
        candidates = jnp.vstack(candidates)
      elif select_candidates == 'fy-nullcline':
        candidates = [self.analyzed_results[key][0] for key in self.analyzed_results.keys()
                      if key.startswith(C.fy_nullcline_points)]
        if len(candidates) == 0:
          raise errors.AnalyzerError(f'No nullcline points are found, please call '
                                     f'".{self.plot_nullcline.__name__}()" first.')
        candidates = jnp.vstack(candidates)
      elif select_candidates == 'nullclines':
        candidates = [self.analyzed_results[key][0] for key in self.analyzed_results.keys()
                      if key.startswith(C.fy_nullcline_points) or key.startswith(C.fy_nullcline_points)]
        if len(candidates) == 0:
          raise errors.AnalyzerError(f'No nullcline points are found, please call '
                                     f'".{self.plot_nullcline.__name__}()" first.')
        candidates = jnp.vstack(candidates)
      elif select_candidates == 'aux_rank':
        candidates, _ = self._get_fp_candidates_by_aux_rank(num_rank=num_rank)
      else:
        raise ValueError

    # get fixed points
    if len(candidates):
      fixed_points, _, _ = self._get_fixed_points(jnp.asarray(candidates),
                                                  tol_aux=tol_aux,
                                                  tol_unique=tol_unique,
                                                  tol_opt_candidate=tol_opt_screen)
      utils.output('I am trying to filter out duplicate fixed points ...')
      fixed_points = np.asarray(fixed_points)
      fixed_points, _ = utils.keep_unique(fixed_points, tolerance=tol_unique)
      utils.output(f'{C.prefix}Found {len(fixed_points)} fixed points.')
    else:
      utils.output(f'{C.prefix}Found no fixed points.')
      return

    # stability analysis
    # ------------------
    container = {a: {'x': [], 'y': []} for a in stability.get_2d_stability_types()}
    for i in range(len(fixed_points)):
      x = fixed_points[i, 0]
      y = fixed_points[i, 1]
      fp_type = stability.stability_analysis(self.F_jacobian(x, y))
      utils.output(f"{C.prefix}#{i + 1} {self.x_var}={x}, {self.y_var}={y} is a {fp_type}.")
      container[fp_type]['x'].append(x)
      container[fp_type]['y'].append(y)

    # visualization
    # -------------
    if with_plot:
      for fp_type, points in container.items():
        if len(points['x']):
          plot_style = stability.plot_scheme[fp_type]
          pyplot.plot(points['x'], points['y'], '.', markersize=20, **plot_style, label=fp_type)
      pyplot.legend()
      if show:
        pyplot.show()

    if with_return:
      return fixed_points

  def plot_trajectory(self, initials, duration, plot_durations=None, axes='v-v',
                      dt=None, show=False, with_plot=True, with_return=False, **kwargs):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple, dict
        The initial value setting of the targets. It can be a tuple/list of floats to specify
        each value of dynamical variables (for example, ``(a, b)``). It can also be a
        tuple/list of tuple to specify multiple initial values (for example,
        ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.

        - It can be a int/float (``t_end``) to specify the same running end time,
        - Or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
          the start and end simulation time.
        - Or, it can be a list of tuple (``[(t1_start, t1_end), (t2_start, t2_end)]``)
          to specify the specific start and end simulation time for each initial value.
    plot_durations : tuple, list, optional
        The duration to plot. It can be a tuple with ``(start, end)``. It can
        also be a list of tuple ``[(start1, end1), (start2, end2)]`` to specify
        the plot duration for each initial value running.
    axes : str
        The axes to plot. It can be:

         - 'v-v': Plot the trajectory in the 'x_var'-'y_var' axis.
         - 't-v': Plot the trajectory in the 'time'-'var' axis.
    show : bool
        Whether show or not.
    """
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am plotting the trajectory ...')

    if axes not in ['v-v', 't-v']:
      raise errors.AnalyzerError(f'Unknown axes "{axes}", only support "v-v" and "t-v".')

    # check the initial values
    initials = utils.check_initials(initials, self.target_var_names)

    # 2. format the running duration
    assert isinstance(duration, (int, float))

    # 3. format the plot duration
    plot_durations = utils.check_plot_durations(plot_durations, duration, initials)

    # 5. run the network
    dt = math.get_dt() if dt is None else dt
    traject_model = utils.TrajectModel(
      initial_vars=initials,
      integrals={self.x_var: self.F_int_x, self.y_var: self.F_int_y},
      dt=dt)
    mon_res = traject_model.run(duration=duration)

    if with_plot:
      # plots
      for i, initial in enumerate(zip(*list(initials.values()))):
        # legend
        legend = f'$traj_{i}$: '
        for j, key in enumerate(self.target_var_names):
          legend += f'{key}={round(float(initial[j]), 4)}, '
        legend = legend[:-2]

        # visualization
        start = int(plot_durations[i][0] / dt)
        end = int(plot_durations[i][1] / dt)
        if axes == 'v-v':
          lines = pyplot.plot(mon_res[self.x_var][start: end, i],
                              mon_res[self.y_var][start: end, i],
                              label=legend, **kwargs)
          utils.add_arrow(lines[0])
        else:
          pyplot.plot(mon_res.ts[start: end],
                      mon_res[self.x_var][start: end, i],
                      label=legend + f', {self.x_var}', **kwargs)
          pyplot.plot(mon_res.ts[start: end],
                      mon_res[self.y_var][start: end, i],
                      label=legend + f', {self.y_var}', **kwargs)

      # visualization of others
      if axes == 'v-v':
        pyplot.xlabel(self.x_var)
        pyplot.ylabel(self.y_var)
        scale = (self.lim_scale - 1.) / 2
        pyplot.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
        pyplot.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
        pyplot.legend()
      else:
        pyplot.legend(title='Initial values')

      if show:
        pyplot.show()

    if with_return:
      return mon_res

  def plot_limit_cycle_by_sim(self, initials, duration, tol=0.01, show=False, dt=None):
    """Plot trajectories according to the settings.

    Parameters
    ----------
    initials : list, tuple
        The initial value setting of the targets.

        - It can be a tuple/list of floats to specify each value of dynamical variables
          (for example, ``(a, b)``).
        - It can also be a tuple/list of tuple to specify multiple initial values (for
          example, ``[(a1, b1), (a2, b2)]``).
    duration : int, float, tuple, list
        The running duration. Same with the ``duration`` in ``NeuGroup.run()``.

        - It can be a int/float (``t_end``) to specify the same running end time,
        - Or it can be a tuple/list of int/float (``(t_start, t_end)``) to specify
          the start and end simulation time.
        - Or, it can be a list of tuple (``[(t1_start, t1_end), (t2_start, t2_end)]``)
          to specify the specific start and end simulation time for each initial value.
    show : bool
        Whether show or not.
    """
    global pyplot
    if pyplot is None: from matplotlib import pyplot
    utils.output('I am plotting the limit cycle ...')

    # 1. format the initial values
    initials = utils.check_initials(initials, self.target_var_names)

    # 2. format the running duration
    assert isinstance(duration, (int, float))

    dt = math.get_dt() if dt is None else dt
    traject_model = utils.TrajectModel(
      initial_vars=initials,
      integrals={self.x_var: self.F_int_x, self.y_var: self.F_int_y},
      dt=dt)
    mon_res = traject_model.run(duration=duration)

    # 5. run the network
    for init_i, initial in enumerate(zip(*list(initials.values()))):
      #   5.2 run the model
      x_data = mon_res[self.x_var][:, init_i]
      y_data = mon_res[self.y_var][:, init_i]
      max_index = utils.find_indexes_of_limit_cycle_max(x_data, tol=tol)
      if max_index[0] != -1:
        x_cycle = x_data[max_index[0]: max_index[1]]
        y_cycle = y_data[max_index[0]: max_index[1]]
        # 5.5 visualization
        lines = pyplot.plot(x_cycle, y_cycle, label='limit cycle')
        utils.add_arrow(lines[0])
      else:
        utils.output(f'No limit cycle found for initial value {initial}')

    # 6. visualization
    pyplot.xlabel(self.x_var)
    pyplot.ylabel(self.y_var)
    scale = (self.lim_scale - 1.) / 2
    pyplot.xlim(*utils.rescale(self.target_vars[self.x_var], scale=scale))
    pyplot.ylim(*utils.rescale(self.target_vars[self.y_var], scale=scale))
    pyplot.legend()

    if show:
      pyplot.show()



# -*- coding: utf-8 -*-

import warnings
from typing import Union

import jax.numpy as jnp

import brainpy.math as bm
from brainpy.dyn.base import NeuGroup
from brainpy.errors import ModelBuildError
from brainpy.initialize import Initializer, init_param
from brainpy.types import Shape

__all__ = [
  'SpikeTimeInput',
  'PoissonInput',
  'SpikeTimeGroup',
  'PoissonGroup',
]


class SpikeTimeGroup(NeuGroup):
  """The input neuron group characterized by spikes emitting at given times.

  >>> # Get 2 neurons, firing spikes at 10 ms and 20 ms.
  >>> SpikeTimeGroup(2, times=[10, 20])
  >>> # or
  >>> # Get 2 neurons, the neuron 0 fires spikes at 10 ms and 20 ms.
  >>> SpikeTimeGroup(2, times=[10, 20], indices=[0, 0])
  >>> # or
  >>> # Get 2 neurons, neuron 0 fires at 10 ms and 30 ms, neuron 1 fires at 20 ms.
  >>> SpikeTimeGroup(2, times=[10, 20, 30], indices=[0, 1, 0])
  >>> # or
  >>> # Get 2 neurons; at 10 ms, neuron 0 fires; at 20 ms, neuron 0 and 1 fire;
  >>> # at 30 ms, neuron 1 fires.
  >>> SpikeTimeGroup(2, times=[10, 20, 20, 30], indices=[0, 0, 1, 1])

  Parameters
  ----------
  size : int, tuple, list
      The neuron group geometry.
  indices : list, tuple, np.ndarray, JaxArray, jax.numpy.ndarray
      The neuron indices at each time point to emit spikes.
  times : list, tuple, np.ndarray, JaxArray, jax.numpy.ndarray
      The time points which generate the spikes.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(
      self,
      size: Shape,
      times,
      indices,
      need_sort: bool = True,
      name: str = None
  ):
    super(SpikeTimeGroup, self).__init__(size=size, name=name)

    # parameters
    if len(indices) != len(times):
      raise ModelBuildError(f'The length of "indices" and "times" must be the same. '
                            f'However, we got {len(indices)} != {len(times)}.')
    self.num_times = len(times)

    # data about times and indices
    self.i = bm.Variable(bm.zeros(1, dtype=bm.int_))
    self.times = bm.Variable(bm.asarray(times, dtype=bm.float_))
    self.indices = bm.Variable(bm.asarray(indices, dtype=bm.int_))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    if need_sort:
      sort_idx = bm.argsort(self.times)
      self.indices.value = self.indices[sort_idx]
      self.times.value = self.times[sort_idx]

    # functions
    def cond_fun(t):
      return bm.logical_and(self.i[0] < self.num_times, t >= self.times[self.i[0]])

    def body_fun(t):
      self.spike[self.indices[self.i[0]]] = True
      self.i[0] += 1

    self._run = bm.make_while(cond_fun, body_fun, dyn_vars=self.vars())

  def update(self, _t, _i, **kwargs):
    self.spike[:] = False
    self._run(_t)


def SpikeTimeInput(*args, **kwargs):
  """Spike Time Input.

  .. deprecated:: 2.1.0
     Please use "brainpy.dyn.SpikeTimeGroup" instead.

  Returns
  -------
  group: NeuGroup
    The neural group.
  """
  warnings.warn('Please use "brainpy.dyn.SpikeTimeGroup" instead. '
                '"brainpy.dyn.SpikeTimeInput" is deprecated since '
                'version 2.1.5', DeprecationWarning)
  return SpikeTimeGroup(*args, **kwargs)


class PoissonGroup(NeuGroup):
  """Poisson Neuron Group.
  """

  def __init__(
      self,
      size: Shape,
      freqs: Union[float, jnp.ndarray, bm.JaxArray, Initializer],
      seed: int = None,
      name: str = None
  ):
    super(PoissonGroup, self).__init__(size=size, name=name)

    self.freqs = init_param(freqs, self.num, allow_none=False)
    self.dt = bm.get_dt() / 1000.
    self.size = (size,) if isinstance(size, int) else tuple(size)
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)
    self.rng = bm.random.RandomState(seed=seed)

  def update(self, _t, _i):
    self.spike.update(self.rng.random(self.num) <= self.freqs * self.dt)
    self.t_last_spike.update(bm.where(self.spike, _t, self.t_last_spike))


def PoissonInput(*args, **kwargs):
  """Poisson Group Input.

  .. deprecated:: 2.1.0
     Please use "brainpy.dyn.PoissonGroup" instead.

  Returns
  -------
  poisson_group: NeuGroup
    The poisson neural group.
  """
  warnings.warn('Please use "brainpy.dyn.PoissonGroup" instead. '
                '"brainpy.dyn.PoissonInput" is deprecated since '
                'version 2.1.5', DeprecationWarning)
  return PoissonGroup(*args, **kwargs)



# -*- coding: utf-8 -*-


"""
This module provides basic tool for graphs, including

- detect the senders and receivers in the network graph,
- find input and output nodes in a given graph,
- detect the cycle in the graph,
- detect the path between two nodes.

"""


from collections import deque, defaultdict

__all__ = [
  'find_senders_and_receivers',
  'find_entries_and_exits',
  'detect_cycle',
  'detect_path',
]


def find_senders_and_receivers(edges):
  """Find all senders and receivers in the given graph."""
  senders = dict()  # find parents according to the child
  receivers = dict()  # find children according to the parent
  for edge in edges:
    sender, receiver = edge
    if receiver not in senders:
      senders[receiver] = [sender]
    else:
      senders[receiver].append(sender)
    if sender not in receivers:
      receivers[sender] = [receiver]
    else:
      receivers[sender].append(receiver)
  return senders, receivers


def find_entries_and_exits(nodes, ff_edges, fb_edges=()):
  """Find input nodes and output nodes."""
  nodes = set(nodes)
  ff_senders = set([n for n, _ in ff_edges])
  ff_receivers = set([n for _, n in ff_edges])
  fb_senders = set([n for n, _ in fb_edges])
  fb_receivers = set([n for _, n in fb_edges])

  # # check lonely feedback nodes
  # fb_receivers_without_ff = fb_receivers - ff_receivers - ff_senders
  # if len(fb_receivers_without_ff) > 0:
  #   raise ValueError(f'Found feedback nodes do not define feedforward connections: \n\n'
  #                    f'{fb_receivers_without_ff}')

  # check lonely nodes
  lonely = nodes - ff_senders - ff_receivers - fb_senders - fb_receivers
  # if len(lonely):
  #   _str_nodes = '\n'.join([str(node) for node in lonely])
  #   raise ValueError(f"Found lonely nodes \n\n{_str_nodes} \n\n"
  #                    f"which do not connect with any other.")

  # get input and output nodes
  entry_points = (ff_senders | fb_senders) - ff_receivers - lonely
  end_points = ff_receivers - ff_senders - lonely
  return list(entry_points), list(end_points)


def topological_sort(nodes, ff_edges, inputs=None):
  if inputs is None:
    inputs, _ = find_entries_and_exits(nodes, ff_edges)
  parents, children = find_senders_and_receivers(ff_edges)
  # using Kahn's algorithm
  ordered_nodes = []
  ff_edges = set(ff_edges)
  inputs = deque(inputs)
  while len(inputs) > 0:
    n = inputs.pop()
    ordered_nodes.append(n)
    for m in children.get(n, ()):
      ff_edges.remove((n, m))
      parents[m].remove(n)
      if parents.get(m) is None or len(parents[m]) < 1:
        inputs.append(m)
  if len(ff_edges) > 0:
    raise RuntimeError("Model has a cycle: impossible "
                       "to automatically determine operation "
                       "order in the model.")
  else:
    return ordered_nodes


def _detect_cycle(v, visited, stacks, graph):
  # visited数组元素为true，标记该元素被isCyclicUtil递归调用链处理中，或处理过
  # recStack数组元素为true，表示该元素还在递归函数isCyclicUtil的函数栈中
  visited[v] = True
  stacks[v] = True
  # 深度遍历所有节点。
  for neighbour in graph[v]:
    if not visited[neighbour]:  # 如果该节点没有被处理过，那么继续调用递归
      if _detect_cycle(neighbour, visited, stacks, graph):  # 如果邻接点neighbour的递归发现了环
        return True  # 那么返回真
    elif stacks[neighbour]:  # 如果neighbour被处理中（这里强调了不是处理过），且还在递归栈中，说明发现了环
      return True
  stacks[v] = False  # 函数开始时，V节点进栈。所以函数结束时，V节点出栈。
  return False  # v的所有邻接点的递归都没有发现环，则返回假


def detect_cycle(nodes, edges):
  """Detect whether a cycle exists in the defined graph.
  """
  node2id = {node: i for i, node in enumerate(nodes)}
  graph = defaultdict(list)
  for s, r in edges:
    graph[node2id[s]].append(node2id[r])
  num = len(nodes)

  visited = [False] * num
  stacks = [False] * num
  for i in range(num):  # 分别以每个节点作为起点，然后开始深度遍历
    if not visited[i]:  # 这里为真，说明之前的深度遍历已经遍历过该节点了，且那次遍历没有发现环
      if _detect_cycle(i, visited, stacks, graph):  # 如果发现环，直接返回
        return True
  return False  # 如果分别以每个节点作为起点的深度遍历都没有发现环，那肯定是整个图没有环


def _has_path_by_dfs(from_node, to_node, graph):
  # queue本质上是堆栈，用来存放需要进行遍历的数据
  # order里面存放的是具体的访问路径
  queue, order = [], []
  # 首先将初始遍历的节点放到queue中，表示将要从这个点开始遍历
  queue.append(from_node)
  while len(queue):
    # 从queue中pop出点v，然后从v点开始遍历了，所以可以将这个点pop出，然后将其放入order中
    # 这里才是最有用的地方，pop（）表示弹出栈顶，由于下面的for循环不断的访问子节点，并将子节点压入堆栈，
    # 也就保证了每次的栈顶弹出的顺序是下面的节点
    v = queue.pop()
    order.append(v)
    # 这里开始遍历v的子节点
    for w in graph[v]:
      # w既不属于queue也不属于order，意味着这个点没被访问过，所以讲起放到queue中，然后后续进行访问
      if w not in order and w not in queue:
        if w == to_node:
          return True
        else:
          queue.append(w)
  return False


def _has_path_by_bfs(from_node, to_node, graph):
  # queue本质上是堆栈，用来存放需要进行遍历的数据
  # order里面存放的是具体的访问路径
  queue, order = [], []
  # 首先将初始遍历的节点放到queue中，表示将要从这个点开始遍历
  # 由于是广度优先，也就是先访问初始节点的所有的子节点，所以可以
  queue.append(from_node)
  order.append(from_node)
  while len(queue):
    # queue.pop(0)意味着是队列的方式出元素，就是先进先出，而下面的for循环将节点v的所有子节点
    # 放到queue中，所以queue.pop(0)就实现了每次访问都是先将元素的子节点访问完毕，而不是优先叶子节点
    v = queue.pop(0)
    for w in graph[v]:
      if w not in order:
        if w == to_node:
          return True
        else:
          # 这里可以直接order.append(w) 因为广度优先就是先访问节点的所有下级子节点，所以可以
          # 将self.sequense[v]的值直接全部先给到order
          order.append(w)
          queue.append(w)
  return False


def detect_path(from_node, to_node, edges, method='dfs'):
  """Detect whether there is a path exist in the defined graph
  from ``from_node`` to ``to_node``. """
  graph = defaultdict(list)
  for s, r in edges:
    graph[s].append(r)
  if method == 'dfs':
    return _has_path_by_dfs(from_node, to_node, graph)
  elif method == 'bfs':
    return _has_path_by_bfs(from_node, to_node, graph)
  else:
    raise ValueError(f'Unknown method {method}')



# -*- coding: utf-8 -*-

import types

from brainpy.base.collector import TensorCollector
from brainpy.errors import MonitorError, RunningError
from brainpy.tools.checking import check_dict_data
from .monitor import Monitor

__all__ = [
  'Runner'
]


class Runner(object):
  """Base Runner.

  Parameters
  ----------
  target: Any
    The target model.
  monitors: None, list of str, tuple of str, Monitor
    Variables to monitor.
  jit: bool
  progress_bar: bool
  dyn_vars: Optional, dict
  numpy_mon_after_run : bool
  """
  def __init__(self, target, monitors=None, fun_monitors=None,
               jit=True, progress_bar=True, dyn_vars=None,
               numpy_mon_after_run=True):
    # target model, while implement __call__ function
    self.target = target

    # jit instruction
    assert isinstance(jit, bool), 'Must be a boolean variable.'
    self.jit = jit

    # monitors
    if monitors is None:
      self.mon = Monitor(variables=[])
    elif isinstance(monitors, (list, tuple, dict)):
      self.mon = Monitor(variables=monitors)
    elif isinstance(monitors, Monitor):
      self.mon = monitors
      self.mon.target = self
    else:
      raise MonitorError(f'"monitors" only supports list/tuple/dict/ '
                         f'instance of Monitor, not {type(monitors)}.')
    self.mon.build()  # build the monitor

    # extra monitors
    if fun_monitors is None:
      fun_monitors = dict()
    check_dict_data(fun_monitors, key_type=str, val_type=types.FunctionType)
    self.fun_monitors = fun_monitors

    # progress bar
    assert isinstance(progress_bar, bool), 'Must be a boolean variable.'
    self.progress_bar = progress_bar
    self._pbar = None

    # dynamical changed variables
    if dyn_vars is None:
      dyn_vars = dict()
    if isinstance(dyn_vars, (list, tuple)):
      dyn_vars = {f'_v{i}': v for i, v in enumerate(dyn_vars)}
    if not isinstance(dyn_vars, dict):
      raise RunningError(f'"dyn_vars" must be a dict, but we got {type(dyn_vars)}')
    self.dyn_vars = TensorCollector(dyn_vars)

    # numpy mon after run
    self.numpy_mon_after_run = numpy_mon_after_run




# -*- coding: utf-8 -*-

"""This module provides basic operations for constructing node graphs.

It supports the following operations:

1. feedforward connection: ">>", ">>="
2. feedback connection: "<<", "<<="
3. merge two nodes: "&", "&="
4. select subsets of one node: "[:]"
5. concatenate a sequence of nodes: "[node1, node2, ...]", "(node1, node2, ...)"
6. wrap a set of nodes: "{node1, node2, ...}"

However, all operations should satisfy the following assumptions:

1. Feedback connection of `(node1, node2)` should have a feedforward path from `node2` to `node1`.
2. Feedforward or feedback connections cannot generate a cycle.
3. Cannot concatenate multiple receiver nodes, e.g., `a >> [b, c]` is forbidden, but `a >> {b, c}`
   is allowed.

"""

from itertools import product
from typing import Union, Sequence, Set

from brainpy.nn import graph_flow
from brainpy.nn.base import Node, Network, FrozenNetwork
from brainpy.nn.datatypes import SingleData
from brainpy.nn.nodes.base import Select, Concat
from brainpy.types import Tensor

__all__ = [
  'ff_connect', 'fb_connect', 'merge', 'select', 'concatenate',
]


def _retrieve_nodes_and_edges(senders: Union[Node, Sequence[Node]],
                              receivers: Union[Node, Sequence[Node]]):
  # check senders
  if isinstance(senders, (tuple, list)):
    senders = [concatenate(senders)]
  elif isinstance(senders, set):
    senders = list(senders)
  elif isinstance(senders, Node):
    senders = [senders]
  else:
    raise TypeError(f"Impossible to send connection from {senders}: it is not "
                    f"a Node or a Network instance.")

  # check receivers
  if isinstance(receivers, (tuple, list)):
    raise TypeError('Cannot concatenate a list/tuple of receivers. '
                    'Please use set to wrap multiple receivers instead.')
  elif isinstance(receivers, set):
    receivers = list(receivers)
  elif isinstance(receivers, Node):
    receivers = [receivers]
  else:
    raise TypeError(f"Impossible to send connection to {receivers}: it is not "
                    f"a Node or a Network instance.")

  # fetch all nodes in two subgraphs
  all_nodes = set()
  for node in senders + receivers:
    if isinstance(node, FrozenNetwork):
      raise TypeError(f"Cannot connect {FrozenNetwork.__name__} to other Nodes.")
    if isinstance(node, Network):
      all_nodes.update(set(node.lnodes))
    elif isinstance(node, Node):
      all_nodes.add(node)
    else:
      raise TypeError(f"Impossible to link nodes: object {node} is neither a "
                      f"'brainpy.rnn.Node' nor a 'brainpy.rnn.Network'.")

  # fetch all feedforward edges in two subgraphs
  all_ff_edges = set()
  for node in senders + receivers:
    if isinstance(node, FrozenNetwork):
      raise TypeError(f"Cannot connect {FrozenNetwork.__name__} to other Nodes.")
    if isinstance(node, Network):
      all_ff_edges.update(set(node.ff_edges))

  # fetch all feedback edges in two subgraphs
  all_fb_edges = set()
  for node in senders + receivers:
    if isinstance(node, FrozenNetwork):
      raise TypeError(f"Cannot connect {FrozenNetwork.__name__} to other Nodes.")
    if isinstance(node, Network):
      all_fb_edges.update(set(node.fb_edges))

  # create edges between output nodes of the
  # subgraph 1 and input nodes of the subgraph 2.
  all_senders = set()
  for node in senders:
    if isinstance(node, Network) and not isinstance(node, FrozenNetwork):
      all_senders.update(node.exit_nodes)
    else:
      all_senders.add(node)
  all_receivers = set()
  for node in receivers:
    if isinstance(node, Network) and not isinstance(node, FrozenNetwork):
      all_receivers.update(node.entry_nodes)
    else:
      all_receivers.add(node)

  return all_nodes, all_ff_edges, all_fb_edges, all_senders, all_receivers


def _reorganize_many2one(ff_edges, fb_edges):
  """Reorganize the many-to-one connections.

  If some node whose "data_type" is :py:class:`brainpy.nn.datatypes.SingleData` receives
  multiple feedforward or feedback connections, we should concatenate all feedforward
  inputs (or feedback inputs) into one instance of :py:class:`brainpy.nn.Concat`, then
  the new Concat instance feeds into this node.

  """
  from brainpy.nn.nodes.base import Concat

  new_nodes = []

  # find parents according to the child
  ff_senders = dict()
  for edge in ff_edges:
    sender, receiver = edge
    if receiver not in ff_senders:
      ff_senders[receiver] = [sender]
    else:
      ff_senders[receiver].append(sender)
  for receiver, senders in ff_senders.items():
    if isinstance(receiver.data_pass, SingleData):
      if len(senders) > 1:
        concat_nodes = [node for node in senders if isinstance(node, Concat)]
        if len(concat_nodes) == 1:
          concat = concat_nodes[0]
          for sender in senders:
            if sender != concat:
              ff_edges.remove((sender, receiver))
              ff_edges.add((sender, concat))
        else:
          concat = Concat()
          for sender in senders:
            ff_edges.remove((sender, receiver))
            ff_edges.add((sender, concat))
          ff_edges.add((concat, receiver))
          new_nodes.append(concat)

  # find parents according to the child
  fb_senders = dict()
  for edge in fb_edges:
    sender, receiver = edge
    if receiver not in fb_senders:
      fb_senders[receiver] = [sender]
    else:
      fb_senders[receiver].append(sender)
  for receiver, senders in fb_senders.items():
    if isinstance(receiver.data_pass, SingleData):
      if len(senders) > 1:
        concat_nodes = [node for node in senders if isinstance(node, Concat)]
        if len(concat_nodes) == 1:
          concat = concat_nodes[0]
          for sender in senders:
            if sender != concat:
              fb_edges.remove((sender, receiver))
              ff_edges.add((sender, concat))
        else:
          concat = Concat()
          for sender in senders:
            fb_edges.remove((sender, receiver))
            ff_edges.add((sender, concat))
          fb_edges.add((concat, receiver))
          new_nodes.append(concat)

  return new_nodes, ff_edges, fb_edges


def merge(
    node: Node,
    *other_nodes: Node,
    inplace: bool = False,
    name: str = None,
    need_detect_cycle=True
) -> Network:
  """Merge different :py:class:`~.Node` or :py:class:`brainpy.nn.base.Network`
  instances into a single :py:class:`brainpy.nn.base.Network` instance.

  :py:class:`~.Node` instances contained in the network to merge will be
  gathered in a single network, along with all previously defined connections
  between them, if they exists.

  You can also perform this operation using the ``&`` operator::

      network = (node1 >> node2) & (node1 >> node3))

  This is equivalent to::

      network = merge((node1 >> node2), (node1 >> node3))

  The inplace operator can also be used::

      network &= other_network

  Parameters
  ----------
  node: Network, Node
      First node or network to merge.
  *other_nodes : Network, Node
      All nodes to merge.
  inplace: bool, default to False
      If `True`, then will update `node` inplace. If `node` is not a Network
      instance, this parameter will causes the function to raise an error.
  name: str, optional
      Name of the resulting Network.
  need_detect_cycle: bool
      Whether need to detect the cycle defined in the graph.

  Returns
  -------
  Network
      A new :py:class:`brainpy.nn.base.Network` instance.
  """
  # checking
  for n in other_nodes + (node,):
    if not isinstance(n, Node):
      raise TypeError(f"Impossible to merge nodes: object {type(n)} is not a Node instance.")

  # get all node and edges
  all_nodes = set()
  all_ff_edges = set()
  all_fb_edges = set()
  for n in other_nodes + (node,):
    if isinstance(n, FrozenNetwork):
      raise TypeError(f'{FrozenNetwork.__name__} cannot merge with other nodes.')
    # fuse models nodes and edges (right side argument)
    if isinstance(n, Network):
      all_nodes |= set(n.lnodes)
      all_ff_edges |= set(n.ff_edges)
      all_fb_edges |= set(n.fb_edges)
    elif isinstance(n, Node):
      all_nodes.add(n)

  # reorganize
  new_nodes, all_ff_edges, all_fb_edges = _reorganize_many2one(all_ff_edges, all_fb_edges)
  all_nodes.update(new_nodes)

  # detect cycles in the graph flow
  all_nodes = tuple(all_nodes)
  all_ff_edges = tuple(all_ff_edges)
  all_fb_edges = tuple(all_fb_edges)
  if need_detect_cycle:
    if graph_flow.detect_cycle(all_nodes, all_ff_edges):
      raise ValueError('We detect cycles in feedforward connections. '
                       'Maybe you should replace some connection with '
                       'as feedback ones.')
    if graph_flow.detect_cycle(all_nodes, all_fb_edges):
      raise ValueError('We detect cycles in feedback connections. ')

  if inplace:
    if not isinstance(node, Network) or isinstance(node, FrozenNetwork):
      raise ValueError(f"Impossible to merge nodes inplace: "
                       f"{node} is not a {Network.__name__} instance.")
    return node.replace_graph(nodes=all_nodes,
                              ff_edges=all_ff_edges,
                              fb_edges=all_fb_edges)

  else:
    return Network(nodes=all_nodes,
                   ff_edges=all_ff_edges,
                   fb_edges=all_fb_edges,
                   name=name)


def ff_connect(
    senders: Union[Node, Sequence[Node], Set[Node]],
    receivers: Union[Node, Set[Node]],
    inplace: bool = False,
    name: str = None,
    need_detect_cycle=True
) -> Network:
  """Connect two sequences of :py:class:`~.Node` instances to form
  a :py:class:`brainpy.nn.base.Network` instance. `senders` output will be used as
  input for `receivers` in the created network. This is similar to a
  function composition operation:

  .. math::

      network(x) = (sender \\circ receiver)(x) = receiver(sender(x))

  You can also perform this operation using the ``>>`` operator::

      network = sender >> receiver

  Or using this function::

      network = ff_connect(sender, receiver)

  - `sender` and `receiver` can also be :py:class:`brainpy.nn.base.Network` instances. In this
    case, the new :py:class:`brainpy.nn.base.Network` created will contain all nodes previously
    contained in all the networks, and link all `node1` outputs to all `node2`
    inputs. This allows to chain the  ``>>`` operator::

      step1 = node0 >> node1  # this is a network
      step2 = step1 >> node2  # this is another

  - `node1` can finally be lists or tuples of nodes. In this
    case, all `node1` outputs will be linked to a :py:class:`~.Concat` node to
    concatenate them, and the :py:class:`~.Concat` node will be linked to all
    `node2` inputs::

      # many-concat-to-one
      network = [node1, node2, ..., node] >> node_out

  - If you do not want to concatenate all input nodes, you can use `set` to
    wrap all input nodes at once. Then, `node2` will receive multiple inputs
    defined in `node1`::

      # many-to-one
      network = {node1, node2, ..., node_N} >> node_out

  - In the case of "one-to-many" feedforward connection, `node2` only support
    a set of node. Using list or tuple to wrap multiple receivers will concatenate
    all nodes in the receiver end. This will cause errors::

      # wrong operation of one-to-many
      network = node_in >> {node1, node2, ..., node_N}

      # correct operation of one-to-many
      network = node_in >> {node1, node2, ..., node_N}

  - "many-to-many" connection is also allowed.

    You can still use the ``>>`` operator in this situation,
    except for many-to-many nodes connections::

      # many-to-many
      {node1, node2, ..., node} >> {node1, node2, ..., node}

  Parameters
  ----------
  senders, receivers : Node, sequence of Node
    Nodes or sequence of nodes to connect feedforward connections.
  inplace: bool
    Whether inplace update the node.
  name: str, optional
    Name for the chaining Network.
  need_detect_cycle: bool
    Whether we need to detect cycles exit in the final network.

  Returns
  -------
  Network
      A :py:class:`brainpy.nn.base.Network` instance chaining the nodes.

  Notes
  -----

  Be careful to how you link the different nodes: `reservoirpy` does not
  allow to have circular dependencies between them::

      network = node1 >> node2  # fine
      network = node1 >> node2 >> node1  # raises! data would flow in
                                       # circles forever...
  """

  all_nodes, all_ff_edges, all_fb_edges, ff_senders, ff_receivers = _retrieve_nodes_and_edges(senders, receivers)
  new_ff_edges = set(product(ff_senders, ff_receivers))

  # all outputs from subgraph 1 are connected to
  # all inputs from subgraph 2.
  all_ff_edges |= new_ff_edges

  # reorganize
  new_nodes, all_ff_edges, all_fb_edges = _reorganize_many2one(all_ff_edges, all_fb_edges)
  all_nodes.update(new_nodes)

  # detect cycles in the graph flow
  all_nodes = tuple(all_nodes)
  all_ff_edges = tuple(all_ff_edges)
  all_fb_edges = tuple(all_fb_edges)
  if need_detect_cycle:
    if graph_flow.detect_cycle(all_nodes, all_ff_edges):
      raise ValueError('We detect cycles in feedforward connections. '
                       'Maybe you should replace some connection with '
                       'as feedback ones.')
    if graph_flow.detect_cycle(all_nodes, all_fb_edges):
      raise ValueError('We detect cycles in feedback connections. ')

  # feedforward
  if inplace:
    if not isinstance(receivers, Network):
      raise TypeError(f'Cannot inplace update the feedback connection of a Node instance: {receivers}')
    if name is not None:
      raise ValueError('Cannot set name when inplace=True.')
    receivers.replace_graph(nodes=all_nodes,
                            ff_edges=all_ff_edges,
                            fb_edges=all_fb_edges)
    return receivers
  else:
    return Network(nodes=all_nodes,
                   ff_edges=all_ff_edges,
                   fb_edges=all_fb_edges,
                   name=name)


def fb_connect(
    senders: Union[Node, Sequence[Node], Set[Node]],
    receivers: Union[Node, Set[Node]],
    inplace: bool = False,
    name: str = None,
    need_detect_cycle=True
) -> Node:
  """Create a feedback connection from ``sender`` node to ``receiver`` node.
  Feedbacks nodes will be called at runtime using data from the previous call.

  You can also perform this operation using the ``<<`` operator.

  Which means that a feedback connection is now created between `node1` and
  `node2`. In other words, the forward function of `node1` depends on the
  previous output of `node2`:

  .. math::
      \\mathrm{node1}(x_t) = \\mathrm{node1}(x_t, \\mathrm{node2}(x_{t - 1}))

  You can also use this function to define feedback::

      node1 = fb_connect(node1, node2)
      # without copy (node1 is the same object throughout)
      node1 = fb_connect(node1, node2, inplace=True, name="n1_copy")

  Parameters
  ----------
  receivers : Node
      Node receiving feedback.
  senders : GenericNode
      Node or Network sending feedback
  inplace : bool, defaults to False
      If `True`, then the function returns a copy of `node`.
  name : str, optional
      Name of the copy of `node` if `inplace` is `True`.
  need_detect_cycle: bool
      Whether we need to detect cycles in the defined network.

  Returns
  -------
  Network
      A network with feedback connections.
  """

  all_nodes, all_ff_edges, all_fb_edges, fb_senders, fb_receivers = _retrieve_nodes_and_edges(senders, receivers)

  # detect whether the node implement its own "init_fb_conn()" function
  for node in fb_receivers:
    if not node.is_feedback_input_supported:
      raise ValueError(f'Establish a feedback connection to \n'
                       f'{node}\n'
                       f'is not allowed. Because this node does not '
                       f'support feedback connections.')

  # detect feedforward cycle
  if need_detect_cycle:
    all_nodes1 = list(all_nodes)
    all_ff_edges1 = tuple(all_ff_edges)
    if graph_flow.detect_cycle(all_nodes1, all_ff_edges1):
      raise ValueError('We detect cycles in feedforward connections. '
                       'Maybe you should replace some connection with '
                       'as feedback ones.')
  # establish feedback connections
  new_fb_edges = set(product(fb_senders, fb_receivers))

  # all outputs from subgraph 1 are connected to
  # all inputs from subgraph 2.
  all_fb_edges |= new_fb_edges

  # reorganize
  new_nodes, all_ff_edges, all_fb_edges = _reorganize_many2one(all_ff_edges, all_fb_edges)
  all_nodes.update(new_nodes)

  # detect cycles in the graph flow
  all_nodes = tuple(all_nodes)
  all_ff_edges = tuple(all_ff_edges)
  all_fb_edges = tuple(all_fb_edges)
  if need_detect_cycle:
    if graph_flow.detect_cycle(all_nodes, all_fb_edges):
      raise ValueError('We detect cycles in feedback connections. ')

  # feedback
  if inplace:
    if not isinstance(receivers, Network):
      raise TypeError(f'Cannot inplace update the feedback connection of a Node instance: {receivers}')
    if name is not None:
      raise ValueError('Cannot set name when inplace=True.')
    receivers.replace_graph(nodes=all_nodes,
                            ff_edges=all_ff_edges,
                            fb_edges=all_fb_edges)
    return receivers
  else:
    return Network(nodes=all_nodes,
                   ff_edges=all_ff_edges,
                   fb_edges=all_fb_edges,
                   name=name)


def select(
    node: Node,
    index: Union[int, Sequence[int], Tensor, slice],
    name: str = None
):
  if isinstance(node, Network) and len(node.exit_nodes) != 1:
    raise ValueError(f'Cannot select subsets of states when Network instance '
                     f'"{node}" has multiple output nodes.')
  return ff_connect(node, Select(index=index), name=name, need_detect_cycle=False)


def concatenate(nodes: Sequence[Node], axis=-1, name=None):
  right = Concat(axis=axis)
  model = Network(name=name)
  for node in nodes:
    if isinstance(node, FrozenNetwork):
      raise ValueError('Cannot concat a Frozen network.')
    if isinstance(node, Network) and len(node.exit_nodes) > 1:
      raise ValueError(f'Cannot concatenate network which has {len(node.exit_nodes)} '
                       f'output nodes with other nodes.')
    model = merge(model,
                  ff_connect(node, right, need_detect_cycle=False),
                  inplace=True,
                  need_detect_cycle=False)
  return model



# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import NeuGroup
from brainpy.initialize import OneInit, Uniform, Initializer, init_param
from brainpy.integrators.joint_eq import JointEq
from brainpy.integrators.ode import odeint
from brainpy.tools.checking import check_initializer
from brainpy.types import Shape, Tensor

__all__ = [
  'HH',
  'MorrisLecar',
]


class HH(NeuGroup):
  r"""Hodgkin–Huxley neuron model.

  **Model Descriptions**

  The Hodgkin-Huxley (HH; Hodgkin & Huxley, 1952) model [1]_ for the generation of
  the nerve action potential is one of the most successful mathematical models of
  a complex biological process that has ever been formulated. The basic concepts
  expressed in the model have proved a valid approach to the study of bio-electrical
  activity from the most primitive single-celled organisms such as *Paramecium*,
  right through to the neurons within our own brains.

  Mathematically, the model is given by,

  .. math::

      C \frac {dV} {dt} = -(\bar{g}_{Na} m^3 h (V &-E_{Na})
      + \bar{g}_K n^4 (V-E_K) + g_{leak} (V - E_{leak})) + I(t)

      \frac {dx} {dt} &= \alpha_x (1-x)  - \beta_x, \quad x\in {\rm{\{m, h, n\}}}

      &\alpha_m(V) = \frac {0.1(V+40)}{1-\exp(\frac{-(V + 40)} {10})}

      &\beta_m(V) = 4.0 \exp(\frac{-(V + 65)} {18})

      &\alpha_h(V) = 0.07 \exp(\frac{-(V+65)}{20})

      &\beta_h(V) = \frac 1 {1 + \exp(\frac{-(V + 35)} {10})}

      &\alpha_n(V) = \frac {0.01(V+55)}{1-\exp(-(V+55)/10)}

      &\beta_n(V) = 0.125 \exp(\frac{-(V + 65)} {80})

  The illustrated example of HH neuron model please see `this notebook <../neurons/HH_model.ipynb>`_.

  The Hodgkin–Huxley model can be thought of as a differential equation system with
  four state variables, :math:`V_{m}(t),n(t),m(t)`, and :math:`h(t)`, that change
  with respect to time :math:`t`. The system is difficult to study because it is a
  nonlinear system and cannot be solved analytically. However, there are many numeric
  methods available to analyze the system. Certain properties and general behaviors,
  such as limit cycles, can be proven to exist.

  *1. Center manifold*

  Because there are four state variables, visualizing the path in phase space can
  be difficult. Usually two variables are chosen, voltage :math:`V_{m}(t)` and the
  potassium gating variable :math:`n(t)`, allowing one to visualize the limit cycle.
  However, one must be careful because this is an ad-hoc method of visualizing the
  4-dimensional system. This does not prove the existence of the limit cycle.

  .. image:: ../../../../_static/Hodgkin_Huxley_Limit_Cycle.png
      :align: center

  A better projection can be constructed from a careful analysis of the Jacobian of
  the system, evaluated at the equilibrium point. Specifically, the eigenvalues of
  the Jacobian are indicative of the center manifold's existence. Likewise, the
  eigenvectors of the Jacobian reveal the center manifold's orientation. The
  Hodgkin–Huxley model has two negative eigenvalues and two complex eigenvalues
  with slightly positive real parts. The eigenvectors associated with the two
  negative eigenvalues will reduce to zero as time :math:`t` increases. The remaining
  two complex eigenvectors define the center manifold. In other words, the
  4-dimensional system collapses onto a 2-dimensional plane. Any solution
  starting off the center manifold will decay towards the *center manifold*.
  Furthermore, the limit cycle is contained on the center manifold.

  *2. Bifurcations*

  If the injected current :math:`I` were used as a bifurcation parameter, then the
  Hodgkin–Huxley model undergoes a Hopf bifurcation. As with most neuronal models,
  increasing the injected current will increase the firing rate of the neuron.
  One consequence of the Hopf bifurcation is that there is a minimum firing rate.
  This means that either the neuron is not firing at all (corresponding to zero
  frequency), or firing at the minimum firing rate. Because of the all-or-none
  principle, there is no smooth increase in action potential amplitude, but
  rather there is a sudden "jump" in amplitude. The resulting transition is
  known as a `canard <http://www.scholarpedia.org/article/Canards>`_.

  .. image:: ../../../../_static/Hodgkins_Huxley_bifurcation_by_I.gif
     :align: center

  The following image shows the bifurcation diagram of the Hodgkin–Huxley model
  as a function of the external drive :math:`I` [3]_. The green lines show the amplitude
  of a stable limit cycle and the blue lines indicate unstable limit-cycle behaviour,
  both born from Hopf bifurcations. The solid red line shows the stable fixed point
  and the black line shows the unstable fixed point.

  .. image:: ../../../../_static/Hodgkin_Huxley_bifurcation.png
     :align: center

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> group = bp.dyn.HH(2)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', 10.))
    >>> runner.run(200.)
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, show=True)

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> group = bp.dyn.HH(2)
    >>>
    >>> I1 = bp.inputs.spike_input(sp_times=[500., 550., 1000, 1030, 1060, 1100, 1200], sp_lens=5, sp_sizes=5., duration=2000, )
    >>> I2 = bp.inputs.spike_input(sp_times=[600.,       900, 950, 1500], sp_lens=5, sp_sizes=5., duration=2000, )
    >>> I1 += bp.math.random.normal(0, 3, size=I1.shape)
    >>> I2 += bp.math.random.normal(0, 3, size=I2.shape)
    >>> I = bp.math.stack((I1, I2), axis=-1)
    >>>
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', I, 'iter'))
    >>> runner.run(2000.)
    >>>
    >>> fig, gs = bp.visualize.get_figure(1, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> plt.plot(runner.mon.ts, runner.mon.V[:, 0])
    >>> plt.plot(runner.mon.ts, runner.mon.V[:, 1] + 130)
    >>> plt.xlim(10, 2000)
    >>> plt.xticks([])
    >>> plt.yticks([])
    >>> plt.show()


  **Model Parameters**

  ============= ============== ======== ====================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------------------
  V_th          20.            mV       the spike threshold.
  C             1.             ufarad   capacitance.
  E_Na          50.            mV       reversal potential of sodium.
  E_K           -77.           mV       reversal potential of potassium.
  E_leak        54.387         mV       reversal potential of unspecific.
  g_Na          120.           msiemens conductance of sodium channel.
  g_K           36.            msiemens conductance of potassium channel.
  g_leak        .03            msiemens conductance of unspecific channels.
  ============= ============== ======== ====================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                        -65         Membrane potential.
  m                        0.05        gating variable of the sodium ion channel.
  n                        0.32        gating variable of the potassium ion channel.
  h                        0.60        gating variable of the sodium ion channel.
  input                     0          External and synaptic input current.
  spike                    False       Flag to mark whether the neuron is spiking.
  t_last_spike       -1e7               Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Hodgkin, Alan L., and Andrew F. Huxley. "A quantitative description
         of membrane current and its application to conduction and excitation
         in nerve." The Journal of physiology 117.4 (1952): 500.
  .. [2] https://en.wikipedia.org/wiki/Hodgkin%E2%80%93Huxley_model
  .. [3] Ashwin, Peter, Stephen Coombes, and Rachel Nicks. "Mathematical
         frameworks for oscillatory network dynamics in neuroscience."
         The Journal of Mathematical Neuroscience 6, no. 1 (2016): 1-92.
  """

  def __init__(
      self,
      size: Shape,
      ENa: Union[float, Tensor, Initializer, Callable] = 50.,
      gNa: Union[float, Tensor, Initializer, Callable] = 120.,
      EK: Union[float, Tensor, Initializer, Callable] = -77.,
      gK: Union[float, Tensor, Initializer, Callable] = 36.,
      EL: Union[float, Tensor, Initializer, Callable] = -54.387,
      gL: Union[float, Tensor, Initializer, Callable] = 0.03,
      V_th: Union[float, Tensor, Initializer, Callable] = 20.,
      C: Union[float, Tensor, Initializer, Callable] = 1.0,
      V_initializer: Union[Initializer, Callable, Tensor] = Uniform(-70, -60.),
      m_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.5),
      h_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.6),
      n_initializer: Union[Initializer, Callable, Tensor] = OneInit(0.32),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(HH, self).__init__(size=size, name=name)

    # parameters
    self.ENa = init_param(ENa, self.num, allow_none=False)
    self.EK = init_param(EK, self.num, allow_none=False)
    self.EL = init_param(EL, self.num, allow_none=False)
    self.gNa = init_param(gNa, self.num, allow_none=False)
    self.gK = init_param(gK, self.num, allow_none=False)
    self.gL = init_param(gL, self.num, allow_none=False)
    self.C = init_param(C, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # variables
    check_initializer(m_initializer, 'm_initializer', allow_none=False)
    check_initializer(h_initializer, 'h_initializer', allow_none=False)
    check_initializer(n_initializer, 'n_initializer', allow_none=False)
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    self.m = bm.Variable(init_param(m_initializer, (self.num,)))
    self.h = bm.Variable(init_param(h_initializer, (self.num,)))
    self.n = bm.Variable(init_param(n_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dm(self, m, t, V):
    alpha = 0.1 * (V + 40) / (1 - bm.exp(-(V + 40) / 10))
    beta = 4.0 * bm.exp(-(V + 65) / 18)
    dmdt = alpha * (1 - m) - beta * m
    return dmdt

  def dh(self, h, t, V):
    alpha = 0.07 * bm.exp(-(V + 65) / 20.)
    beta = 1 / (1 + bm.exp(-(V + 35) / 10))
    dhdt = alpha * (1 - h) - beta * h
    return dhdt

  def dn(self, n, t, V):
    alpha = 0.01 * (V + 55) / (1 - bm.exp(-(V + 55) / 10))
    beta = 0.125 * bm.exp(-(V + 65) / 80)
    dndt = alpha * (1 - n) - beta * n
    return dndt

  def dV(self, V, t, m, h, n, I_ext):
    I_Na = (self.gNa * m ** 3.0 * h) * (V - self.ENa)
    I_K = (self.gK * n ** 4.0) * (V - self.EK)
    I_leak = self.gL * (V - self.EL)
    dVdt = (- I_Na - I_K - I_leak + I_ext) / self.C
    return dVdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dm, self.dh, self.dn])

  def update(self, _t, _dt):
    V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input, dt=_dt)
    self.spike.value = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    self.t_last_spike.value = bm.where(self.spike, _t, self.t_last_spike)
    self.V.value = V
    self.m.value = m
    self.h.value = h
    self.n.value = n
    self.input[:] = 0.


class MorrisLecar(NeuGroup):
  r"""The Morris-Lecar neuron model.

  **Model Descriptions**

  The Morris-Lecar model [1]_ (Also known as :math:`I_{Ca}+I_K`-model)
  is a two-dimensional "reduced" excitation model applicable to
  systems having two non-inactivating voltage-sensitive conductances.
  This model was named after Cathy Morris and Harold Lecar, who
  derived it in 1981. Because it is two-dimensional, the Morris-Lecar
  model is one of the favorite conductance-based models in computational neuroscience.

  The original form of the model employed an instantaneously
  responding voltage-sensitive Ca2+ conductance for excitation and a delayed
  voltage-dependent K+ conductance for recovery. The equations of the model are:

  .. math::

      \begin{aligned}
      C\frac{dV}{dt} =& -  g_{Ca} M_{\infty} (V - V_{Ca})- g_{K} W(V - V_{K}) -
                        g_{Leak} (V - V_{Leak}) + I_{ext} \\
      \frac{dW}{dt} =& \frac{W_{\infty}(V) - W}{ \tau_W(V)}
      \end{aligned}

  Here, :math:`V` is the membrane potential, :math:`W` is the "recovery variable",
  which is almost invariably the normalized :math:`K^+`-ion conductance, and
  :math:`I_{ext}` is the applied current stimulus.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>>
    >>> group = bp.dyn.MorrisLecar(1)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V', 'W'], inputs=('input', 100.))
    >>> runner.run(1000)
    >>>
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.W, ylabel='W')
    >>> fig.add_subplot(gs[1, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, ylabel='V', show=True)


  **Model Parameters**

  ============= ============== ======== =======================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- -------------------------------------------------------
  V_Ca          130            mV       Equilibrium potentials of Ca+.(mV)
  g_Ca          4.4            \        Maximum conductance of corresponding Ca+.(mS/cm2)
  V_K           -84            mV       Equilibrium potentials of K+.(mV)
  g_K           8              \        Maximum conductance of corresponding K+.(mS/cm2)
  V_Leak        -60            mV       Equilibrium potentials of leak current.(mV)
  g_Leak        2              \        Maximum conductance of leak current.(mS/cm2)
  C             20             \        Membrane capacitance.(uF/cm2)
  V1            -1.2           \        Potential at which M_inf = 0.5.(mV)
  V2            18             \        Reciprocal of slope of voltage dependence of M_inf.(mV)
  V3            2              \        Potential at which W_inf = 0.5.(mV)
  V4            30             \        Reciprocal of slope of voltage dependence of W_inf.(mV)
  phi           0.04           \        A temperature factor. (1/s)
  V_th          10             mV       The spike threshold.
  ============= ============== ======== =======================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                  -20               Membrane potential.
  W                  0.02              Gating variable, refers to the fraction of
                                       opened K+ channels.
  input              0                 External and synaptic input current.
  spike              False             Flag to mark whether the neuron is spiking.
  t_last_spike       -1e7              Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Meier, Stephen R., Jarrett L. Lancaster, and Joseph M. Starobin.
         "Bursting regimes in a reaction-diffusion system with action
         potential-dependent equilibrium." PloS one 10.3 (2015):
         e0122401.
  .. [2] http://www.scholarpedia.org/article/Morris-Lecar_model
  .. [3] https://en.wikipedia.org/wiki/Morris%E2%80%93Lecar_model
  """

  def __init__(
      self,
      size: Shape,
      V_Ca: Union[float, Tensor, Initializer, Callable] = 130.,
      g_Ca: Union[float, Tensor, Initializer, Callable] = 4.4,
      V_K: Union[float, Tensor, Initializer, Callable] = -84.,
      g_K: Union[float, Tensor, Initializer, Callable] = 8.,
      V_leak: Union[float, Tensor, Initializer, Callable] = -60.,
      g_leak: Union[float, Tensor, Initializer, Callable] = 2.,
      C: Union[float, Tensor, Initializer, Callable] = 20.,
      V1: Union[float, Tensor, Initializer, Callable] = -1.2,
      V2: Union[float, Tensor, Initializer, Callable] = 18.,
      V3: Union[float, Tensor, Initializer, Callable] = 2.,
      V4: Union[float, Tensor, Initializer, Callable] = 30.,
      phi: Union[float, Tensor, Initializer, Callable] = 0.04,
      V_th: Union[float, Tensor, Initializer, Callable] = 10.,
      W_initializer: Union[Callable, Initializer, Tensor] = OneInit(0.02),
      V_initializer: Union[Callable, Initializer, Tensor] = Uniform(-70., -60.),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(MorrisLecar, self).__init__(size=size, name=name)

    # params
    self.V_Ca = init_param(V_Ca, self.num, allow_none=False)
    self.g_Ca = init_param(g_Ca, self.num, allow_none=False)
    self.V_K = init_param(V_K, self.num, allow_none=False)
    self.g_K = init_param(g_K, self.num, allow_none=False)
    self.V_leak = init_param(V_leak, self.num, allow_none=False)
    self.g_leak = init_param(g_leak, self.num, allow_none=False)
    self.C = init_param(C, self.num, allow_none=False)
    self.V1 = init_param(V1, self.num, allow_none=False)
    self.V2 = init_param(V2, self.num, allow_none=False)
    self.V3 = init_param(V3, self.num, allow_none=False)
    self.V4 = init_param(V4, self.num, allow_none=False)
    self.phi = init_param(phi, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)

    # vars
    check_initializer(V_initializer, 'V_initializer', allow_none=False)
    check_initializer(W_initializer, 'W_initializer', allow_none=False)
    self.W = bm.Variable(init_param(W_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dV(self, V, t, W, I_ext):
    M_inf = (1 / 2) * (1 + bm.tanh((V - self.V1) / self.V2))
    I_Ca = self.g_Ca * M_inf * (V - self.V_Ca)
    I_K = self.g_K * W * (V - self.V_K)
    I_Leak = self.g_leak * (V - self.V_leak)
    dVdt = (- I_Ca - I_K - I_Leak + I_ext) / self.C
    return dVdt

  def dW(self, W, t, V):
    tau_W = 1 / (self.phi * bm.cosh((V - self.V3) / (2 * self.V4)))
    W_inf = (1 / 2) * (1 + bm.tanh((V - self.V3) / self.V4))
    dWdt = (W_inf - W) / tau_W
    return dWdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dW])

  def update(self, _t, _dt):
    V, self.W.value = self.integral(self.V, self.W, _t, self.input, dt=_dt)
    spike = bm.logical_and(self.V < self.V_th, V >= self.V_th)
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = V
    self.spike.value = spike
    self.input[:] = 0.
